﻿#pragma TextEncoding = "UTF-8"
#pragma rtGlobals=3
#pragma DefaultTab={3,20,4}
#pragma ModuleName=Hyperspec
#pragma IgorVersion=8
#pragma version=1.03

#include <Resize Controls>

// Project Updater header
static constant kProjectID = 22408 // the project node on IgorExchange
static strconstant ksShortTitle = "Hyperspec" // the project short title on IgorExchange

//#define dynamic

static strconstant ksPackageName   = "Hyperspec"
static strconstant ksPrefsFileName = "acwHyperspec.bin"
static constant    kPrefsVersion   = 100

menu "New"
	"Hyperspectral browser...", /Q, Hyperspec#NewHyperspecBrowser()
end

#if IgorVersion() >= 9
#ifdef dynamic
menu "DataBrowserObjectsPopup", dynamic
	Hyperspec#ValidBrowserWave("Hyperspectral browser", 1, 1), /Q, Hyperspec#PlotDataBrowserSelection()
end
#else
menu "DataBrowserObjectsPopup"
	"Hyperspectral browser", /Q, Hyperspec#PlotDataBrowserSelection()
end
#endif
#endif

static structure PackagePrefs
	uint16 version
	uint16 options
	uint16 ctnum
	
	STRUCT RGBColor chrgb  // 6 bytes
	STRUCT point crosshair // 4 bytes
	uint32 plane // 4
	float  imin
	float  imax
	char   restart
	char   unit[15]  // text
	
	char   reserved[128 - 6 - 6 - 4 - 12 - 1 - 15] // 44 bytes used
endstructure

static function PrefsSetDefaults(STRUCT PackagePrefs &prefs)
	prefs.version  = kPrefsVersion
	prefs.options  = 0 // 1: reverse spectrum, 2: reverse image y, 4: log colors; 8: manual start index, 16: manual end index
	                   // bits 0, 1, 2 are retained. bits 3 & 4 are set to 0 unless we are rebuilding the panel.
	prefs.unit     = "wavenumber"
	prefs.ctnum    = 2 // yellowhot
	prefs.chrgb.red   = 65535
	prefs.chrgb.green = 65535
	prefs.chrgb.blue  = 65535
	
	int i
	for(i=0;i<(128-44);i+=1)
		prefs.reserved[i] = 0
	endfor
end

static function PrefsUpdate(STRUCT PackagePrefs &prefs)
	if (prefs.version < 100)
//		PrefsSetDefaults(prefs)
		prefs.version = kPrefsVersion
	endif
end

static function PrefsLoad(STRUCT PackagePrefs &prefs)
	LoadPackagePreferences/MIS=1 ksPackageName, ksPrefsFileName, 0, prefs
	if (V_flag!=0 || V_bytesRead==0)
		PrefsSetDefaults(prefs)
	elseif (prefs.version != kPrefsVersion)
		// prefs definition may be changed
		PrefsUpdate(prefs)
	endif
	prefs.ctnum    = limit(prefs.ctnum, 0, ItemsInList(CTabList())-1)
end

static function PrefsSave()
	
	if (str2num(GetUserData("HyperspecBrowser", "", "version")) != GetThisVersion())
		return 0
	endif
	
	STRUCT PackagePrefs prefs
	PrefsLoad(prefs)
	
	ControlInfo/W=HyperspecBrowser popCT
	prefs.ctnum = V_Value - 1

	ControlInfo/W=HyperspecBrowser popColor
	prefs.chrgb.red   = V_Red
	prefs.chrgb.green = V_Green
	prefs.chrgb.blue  = V_Blue
	
	// 1: reverse spectrum, 2: reverse image y, 4: log colors; 8: manual start index, 16: manual end index
	
	ControlInfo/W=HyperspecBrowser chkLog
	prefs.options = setbitvalue(prefs.options, 2, v_value)
	ControlInfo/W=HyperspecBrowser chkAutoMin
	prefs.options = setbitvalue(prefs.options, 3, !v_value)
	ControlInfo/W=HyperspecBrowser chkAutoMax
	prefs.options = setbitvalue(prefs.options, 4, !v_value)
	ControlInfo/W=HyperspecBrowser svMin
	prefs.imin = v_value
	ControlInfo/W=HyperspecBrowser svMax
	prefs.imax = v_value
	ControlInfo/W=HyperspecBrowser svPlane
	prefs.plane = v_value
	
	prefs.crosshair.h = pcsr(A, "HyperspecBrowser#G0")
	prefs.crosshair.v = qcsr(A, "HyperspecBrowser#G0")

	SavePackagePreferences ksPackageName, ksPrefsFileName, 0, prefs
end

static function setbitvalue(int var, int bit, int value)
	return (var&~2^bit) | value*2^bit
end

static function MakePrefsPanel()
	STRUCT PackagePrefs prefs
	PrefsLoad(prefs)
	
	variable WL = 150, WT = 100, height = 130, width = 260 // window coordinates
	GetWindow/Z HyperspecBrowser wsizeRM
	if (v_flag == 0)
		WL = V_right; WT = V_top
	endif
	NewPanel/K=1/N=HyperspecPrefsPanel/W=(WL,WT,WL+width,WT+height) as "Hyperspec Settings"
	ModifyPanel/W=HyperspecPrefsPanel, fixedSize=1, noEdit=1
	
	SetVariable svUnit, pos={10,20}, size={240,16}, title="Spectrum x-axis unit"//, Proc=hyperspec#BaselineSetvars
	SetVariable svUnit, help={"Unit for spectrum independent variable"}, fSize=12, value=_STR:prefs.unit
	CheckBox chkRevSpectrum pos={10,50},size={100,16},title="Reverse spectrum x-axis",fSize=12,value=prefs.options&1
	CheckBox chkRevY pos={10,70},size={100,16},title="Reverse image y-axis",fSize=12,value=(prefs.options&2)>0
	
	Button btnCancel, pos={10,100}, size={60,22}, title="Cancel", Proc=hyperspec#ButtonProc
	Button btnSave, pos={200,100}, size={50,22}, title="Save", Proc=hyperspec#ButtonProc
	
	PauseForUser HyperspecPrefsPanel, HyperspecBrowser
end

// called when prefs panel is closed
static function PrefsClosePanel(int DoIt)
	
	STRUCT PackagePrefs prefs
	PrefsLoad(prefs)
	prefs.restart = 0
	
	if (doit)
				
		int oldoptions = prefs.options
		
		ControlInfo/W=HyperspecPrefsPanel svUnit
		if (cmpstr(s_value, prefs.unit))
			prefs.unit = s_value
			prefs.restart = 1
		endif
				
		ControlInfo/W=HyperspecPrefsPanel chkRevSpectrum
		prefs.options = setbitvalue(prefs.options, 0, v_value)
		ControlInfo/W=HyperspecPrefsPanel chkRevY
		prefs.options = setbitvalue(prefs.options, 1, v_value)
		
		if (oldoptions != prefs.options)
			prefs.restart = 1
		endif
		
		SavePackagePreferences ksPackageName, ksPrefsFileName, 0, prefs
	endif
	
	// kill the preferences panel
	KillWindow/Z HyperspecPrefsPanel
	if (prefs.restart)
		MakeHyperspecBrowser(GetImageWave())
	endif
		
end

// returns truth that this procedure file has been updated since initialisation
static function CheckUpdated(int restart)
	if (str2num(GetUserData("HyperspecBrowser", "", "version")) != GetThisVersion())
		if (restart)
			DoAlert 0, "You have updated the package since this panel was created.\r\rThe package will restart to update the control panel."
			MakeHyperspecBrowser(GetImageWave())
		else
			DoAlert 0, "You have updated the package since this panel was created.\r\rPlease restart to continue."
		endif
		return 1
	endif
	return 0
end

static function demo()
	Make/O/N=(64,64,501) wDemo3d
	// make a broad Gaussian peak that increases in intensity towards the centre of the field of view
	wDemo3d  = 1000 * (60-sqrt((x-32)^2 + (y-32)^2)) * Gauss(z,350,400)
	// add a linear baseline whose slope varies depending on position
	wDemo3d += (0.0002 * x + 0.0005 * y) * z
	// add a peak at position 200
	wDemo3d += 500 * Gauss(z,200,20)
	// add a peak at position 150 to spectra that fall within a circular area
	wDemo3d += sqrt((x-20)^2 + (y-20)^2) < 10 ? 1000 * Gauss(z,150,20) : 0
	// add some noise
	wDemo3d += gnoise(1)
	MakeHyperspecBrowser(wDemo3d)
end

static function/S ValidBrowserWave(string str, int selectOne, int datacube)
	if (strlen(GetBrowserSelection(-1)) == 0 || (strlen(GetBrowserSelection(1)) && selectOne))
		return "" // Data Browser is not open/multiple objects selected
	endif
	wave/Z w1 = $GetBrowserSelection(0)
	if (WaveExists(w1) && WaveType(w1,1)==1)
		if (datacube == 0)
			return str
		elseif (WaveDims(w1)==3)
			return str
		endif
	endif
	return ""
end

static function Valid3DWave(wave w)
	return WaveExists(w) && WaveType(w,1)==1 && WaveDims(w)==3
end

static function PlotDataBrowserSelection()
	wave/Z w = $GetBrowserSelection(0)
	if (WaveExists(w) && WaveDims(w)==3)
		MakeHyperspecBrowser(w)
	endif
end

static function/DF getDFR()
	DFREF dfr = root:Packages:Hyperspec
	if (DataFolderRefStatus(dfr) != 1)
		return ResetDFR()
	endif
	return dfr
end

static function/DF ResetDFR()
	NewDataFolder/O root:Packages
	NewDataFolder/O root:Packages:Hyperspec
	DFREF dfr=root:Packages:Hyperspec
	return dfr
end

static function NewHyperspecBrowser()
	CreateBrowser/M Prompt="Select a 3D wave that represents a hyperspectral datacube"
	ModifyBrowser/M showWaves=1, showVars=0, showStrs=0, ShowInfo=0, showPlot=1
	ModifyBrowser/M showModalBrowser
	if (ItemsInList(S_BrowserList) == 1)
		wave/Z wSel = $StringFromList(0, S_BrowserList)
		MakeHyperspecBrowser(wSel)
	endif
	return 0
end

static function MakeHyperspecBrowser(wave/Z w3D)
	
	if (!Valid3DWave(w3D))
		NewHyperspecBrowser()
		return 0
	endif
		
	DFREF dfr = GetDFR()
	Make/O/N=(DimSize(w3D, 0),DimSize(w3D, 1)) dfr:wMask
	Make/O/N=(DimSize(w3D, 0),DimSize(w3D, 1),4) dfr:wDisplayMask/wave=wDisplayMask
	CopyScales w3D, wDisplayMask
	
	Make/O/N=(DimSize(w3D, 2))/Y=(WaveType(w3D)) dfr:spectrum/wave=spectrum
	SetScale/P x, DimOffset(w3D, 2), DimDelta(w3D, 2), spectrum
	wDisplayMask[][][0,2] = 65535
	wDisplayMask[][][3] = 0
	
	variable left, top, width, height, oldright, oldbottom
	
	width = 385
	height = 675
	
	GetWindow/Z HyperspecBrowser wsizeRM // points
	if (V_flag == 0)
		left = v_left
		top = v_top
		oldright = v_right
		oldbottom = v_bottom
	else
		GetMouse
		left = v_left - 100
		top = v_top - 200
	endif
	
//	if (!prefs.restart)
//		KillWindow/Z HyperspecBrowser
//	else
//		PrefsSave()
//	endif
	
	KillWindow/Z HyperspecBrowser
		
	STRUCT PackagePrefs prefs
	PrefsLoad(prefs)
	
	string strColorTable = StringFromList(prefs.ctnum, CTabList())
	int logScale = prefs.options & 4 ? 1 : 0
	
	variable planeStart = IndexToScale(w3D,0,2)
	variable planeEnd = IndexToScale(w3D,DimSize(w3D,2)-1,2)
	variable planeMin =  min(planeStart, planeEnd)
	variable planeMax =  max(planeStart, planeEnd)
	
	if (!prefs.restart || prefs.plane<planeMin || prefs.plane>planeMax)
		prefs.plane = (planeMin + planeMax) / 2
	endif

	variable layer = scaleToIndex(w3D, prefs.plane, 2)
			
	int xpnt, ypnt
	xpnt = DimSize(w3D, 0) / 2
	ypnt = DimSize(w3D, 1) / 2
	
	if (prefs.restart)
		xpnt = prefs.crosshair.h
		ypnt = prefs.crosshair.v
		if (!prefs.restart || xpnt != limit(xpnt, 0, DimSize(w3D, 0)-1))
			xpnt = DimSize(w3D, 0) / 2
		endif
		if (ypnt != limit(ypnt, 0, DimSize(w3D, 1)-1))
			ypnt = DimSize(w3D, 1) / 2
		endif
	endif
	
	if (!prefs.restart) // autoscale colors
		prefs.options = setbitvalue(prefs.options, 3, 0)
		prefs.options = setbitvalue(prefs.options, 4, 0)
	endif
	
	variable vMin, vMax
	WaveStats/Q/M=1/RMD=[][][layer] w3D
	vMin = prefs.options&8 ? prefs.imin : v_min
	vMax = prefs.options&16 ? prefs.imax : v_max
		
	spectrum = w3D[xpnt][ypnt][p]
	string strTextBox = ""
	
//	NewPanel/W=(left,top,left+width,top+height)/K=1/N=HyperspecBrowser as "Hyperspectral Browser"
	
//	if (!prefs.restart)
	NewPanel/W=(point2cpu(left),point2cpu(top),point2cpu(left+width),point2cpu(top+height))/K=1/N=HyperspecBrowser as "Hyperspectral Browser"
	DoIgorMenu "control", "retrieve window"
//	endif
		
	ModifyPanel/W=HyperspecBrowser cbRGB=(65535,65535,65534)
	
	SetVariable svDatacube,win=HyperspecBrowser,pos={5,5},size={330,14},title="",value=_STR:GetWavesDataFolder(w3D, 2), Proc=Hyperspec#SetvariableProc
	SetVariable svDatacube,win=HyperspecBrowser, help={"Select a 3D wave that represents a hyperspectral datacube"}
	SetVariable svDatacube,win=HyperspecBrowser, userdata=GetWavesDataFolder(w3D, 2)
	
	Button btnInfo, win=HyperspecBrowser, pos={345,5}, size={15,15}, Proc=hyperspec#ButtonProc, title=""
	Button btnInfo, win=HyperspecBrowser, help={"Show Info"}, Picture=hyperspec#pInfo
	
	Button btnSettings, win=HyperspecBrowser, pos={365,5}, size={15,15}, Proc=hyperspec#ButtonProc, title=""
	Button btnSettings, win=HyperspecBrowser, help={"Change Settings"}, Picture=hyperspec#pCog
	
	DefineGuide/W=HyperspecBrowser FGT={FT,30}, FGM={FT,0.27,FB},FGB={FB,-150}//,FGR={FL,0.9,FR}
	Display/W=(21,178,447,433)/FG=(FL,FGM,FR,FGB)/HOST=HyperspecBrowser/N=G0
	
	AppendImage/W=HyperspecBrowser#G0 w3D
	ModifyImage/W=HyperspecBrowser#G0 ''#0 ctab={*,*,$strColorTable,0}, plane=layer, ctabAutoscale=2
	ModifyGraph/W=HyperspecBrowser#G0 margin(left)=8,margin(bottom)=1,margin(right)=38 // or zero for auto margin, allows more space for external color scale
	ModifyGraph/W=HyperspecBrowser#G0 tick=3,mirror=1,noLabel=2,standoff=0
	AppendImage/W=HyperspecBrowser#G0 wDisplayMask
	ColorScale/W=HyperspecBrowser#G0/C/N=scale/F=0/A=RC/X=0/Y=0.00/E/B=1 image=$NameOfWave(w3D), heightPct=100, width=10, tickLen=4, minor=1
	
	if (prefs.options & 2) // reverse image Y
		SetAxis/A/R/W=HyperspecBrowser#G0 left
	endif
	
	// crosshair cursor
	Cursor/W=HyperspecBrowser#G0/P/I/H=1/C=(prefs.chrgb.red,prefs.chrgb.green,prefs.chrgb.blue)/N=1 A $NameOfWave(w3D) xpnt, ypnt
	sprintf strTextBox, "%s[][][%g]", NameOfWave(w3D), layer
	TextBox/C/N=text0/F=0/A=LT/X=8/Y=0/E=2 strTextBox
	
	// rather than plotting w3D[xpnt][ypnt][*] we plot a 1D wave so that it can be easily exported
	Display/W=(39,32,319,116)/FG=(FL,FGT,FR,FGM)/HOST=HyperspecBrowser/N=G1 spectrum/TN=spectrum
	ModifyGraph/W=HyperspecBrowser#G1 margin(left)=8,margin(right)=8
	ModifyGraph/W=HyperspecBrowser#G1 tick(left)=3,noLabel(left)=2,standoff=0,axThick(left)=0
	if (prefs.options & 1)
		SetAxis/A/R/W=HyperspecBrowser#G1 bottom
	else
		SetAxis/A/W=HyperspecBrowser#G1 bottom
	endif
	sprintf strTextBox, "%s[%g][%g]", NameOfWave(w3D), xpnt, ypnt
	TextBox/C/N=text1/F=0/A=LT/X=8/Y=0/E=2 strTextBox
	
		
	ModifyGraph/W=HyperspecBrowser#G1 zColor(spectrum)={spectrum,vMin,vMax,$strColorTable,0}
	SetActiveSubwindow ##
	SetWindow HyperspecBrowser activeChildFrame=0
		
	// vertical hair cursor
	DoUpdate
	GetAxis/W=HyperspecBrowser#G1/Q bottom
	Cursor/W=HyperspecBrowser#G1/P/F/S=2/H=2/N=1 B spectrum (prefs.plane-V_min)/(V_max-V_min), 0.5
		
	variable x0 = 28, y0 = 537
	
	SetVariable svPlane,win=HyperspecBrowser,pos={x0, y0},size={195,180}, value=_NUM:prefs.plane,Proc=Hyperspec#SetvariableProc
	SetVariable svPlane,win=HyperspecBrowser,fSize=12,limits={planeMin,planeMax,abs(DimDelta(w3D,2))}
	SetVariable svPlane,win=HyperspecBrowser,userdata="0" // userdata = "1" to flag updates from cursor hook and avoid reentry
	if (strlen(prefs.unit))
		SetVariable svPlane,win=HyperspecBrowser,title="Plane (" + prefs.unit + ")"
	else
		SetVariable svPlane,win=HyperspecBrowser,title="Layer"
	endif
	
	y0 += 25
	PopupMenu popCT,win=HyperspecBrowser,pos={x0,y0},size={160,20}, Proc=Hyperspec#PopupProc
	PopupMenu popCT,win=HyperspecBrowser,mode=prefs.ctnum+1,value=#"\"*COLORTABLEPOPNONAMES*\"", bodyWidth=160
	CheckBox chkLog,pos={200,y0},size={42,16},title="log scale",fSize=12,value=logScale,Proc=Hyperspec#checkBoxProc
	
	y0 += 20
	CheckBox chkAutoMin,pos={200,y0},size={42,16},title="auto",fSize=12,value=!(prefs.options&8),Proc=Hyperspec#checkBoxProc
	SetVariable svMin,pos={32,y0},size={160,18},bodyWidth=70,value=_NUM:vMin, disable=prefs.options&8 ? 0 : 2
	SetVariable svMin,title="First color index",fSize=12,limits={-Inf,Inf,0},Proc=Hyperspec#SetvariableProc
	
	y0 += 20
	Slider sliderMin pos={x0,y0},size={230,30}, vert=0,side=0,limits={vMin,vMax,0},ticks=0,value=vMin, disable=prefs.options&8 ? 0 : 2, Proc=Hyperspec#SliderProc
	
	y0 += 20
	CheckBox chkAutoMax,pos={200,y0},size={42,16},title="auto",fSize=12,value=!(prefs.options&16),Proc=Hyperspec#checkBoxProc
	SetVariable svMax,pos={32,y0},size={160,18},bodyWidth=70,value=_NUM:vMax, disable=prefs.options&16 ? 0 : 2
	SetVariable svMax,title="Last color index ",fSize=12,limits={-Inf,Inf,0},Proc=Hyperspec#SetvariableProc
	
	y0 += 20
	Slider sliderMax pos={x0,y0},size={230,30}, vert=0,side=0,limits={vMin,vMax,0},ticks=0,value=vMax, disable=prefs.options&16 ? 0 : 2, Proc=Hyperspec#SliderProc
		
	y0 = 537
	x0 = 274
	GroupBox grpPixel,pos={x0,y0},size={86,125},title="Crosshair",fsize=12
	y0 += 22
	x0 += 17
	PopupMenu popColor,win=HyperspecBrowser,pos={x0,y0},size={50,20},mode=3,value=#"\"*COLORPOP*\"", Proc=Hyperspec#PopupProc
	PopupMenu popColor,win=HyperspecBrowser,title="",fsize=12,popColor=(prefs.chrgb.red,prefs.chrgb.green,prefs.chrgb.blue)
	y0 += 22
	Button btnZap,pos={x0,y0},size={50.00,22.00},Proc=Hyperspec#buttonproc
	Button btnZap,title="Zap"
	y0 += 25
	Button btnCopy,pos={x0,y0},size={50.00,22.00},Proc=Hyperspec#buttonproc
	Button btnCopy,title="Copy"
	y0 += 25
	Button btnPaste,pos={x0,y0},size={50.00,22.00},Proc=Hyperspec#buttonproc
	Button btnPaste,title="Paste"
	
	// reset vertical hair cursor position
	GetAxis/W=HyperspecBrowser#G1/Q bottom
	Cursor/W=HyperspecBrowser#G1/F/P B spectrum (prefs.plane-V_min)/(V_max-V_min), 0.5
	logcolors((prefs.options&4) > 0)
	
	ResetPlots(3)
	SetWindow HyperspecBrowser hook(hHyperspec)=Hyperspec#hookHyperspec
	
	Button btnSettings,userdata(ResizeControlsInfo)=A"!!,HqJ,hj-!!#<(!!#<(z!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	Button btnSettings,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#u:Duafnzzzzzzzzzzz"
	Button btnSettings,userdata(ResizeControlsInfo)+=A"zzz!!#u:Duafnzzzzzzzzzzzzzz!!!"
	SetVariable svDatacube,userdata(ResizeControlsInfo)=A"!!,?X!!#9W!!#B_!!#;mz!!#](Aon\"Qzzzzzzzzzzzzzz!!#o2B4uAezz"
	SetVariable svDatacube,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#u:Du]k<zzzzzzzzzzz"
	SetVariable svDatacube,userdata(ResizeControlsInfo)+=A"zzz!!#u:Du]k<zzzzzzzzzzzzzz!!!"
	Button btnInfo,userdata(ResizeControlsInfo)=A"!!,HgJ,hj-!!#<(!!#<(z!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	Button btnInfo,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#u:Du]k<zzzzzzzzzzz"
	Button btnInfo,userdata(ResizeControlsInfo)+=A"zzz!!#u:Du]k<zzzzzzzzzzzzzz!!!"
	SetVariable svPlane,userdata(ResizeControlsInfo)=A"!!,CD!!#Ck5QF/=!!#<Hz!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	SetVariable svPlane,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	SetVariable svPlane,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	PopupMenu popCT,userdata(ResizeControlsInfo)=A"!!,CD!!#CqJ,hqZ!!#<Xz!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	PopupMenu popCT,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	PopupMenu popCT,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	CheckBox chkLog,userdata(ResizeControlsInfo)=A"!!,GX!!#CqJ,hoj!!#<8z!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	CheckBox chkLog,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	CheckBox chkLog,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	CheckBox chkAutoMin,userdata(ResizeControlsInfo)=A"!!,GX!!#D!J,hna!!#<8z!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	CheckBox chkAutoMin,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	CheckBox chkAutoMin,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	SetVariable svMin,userdata(ResizeControlsInfo)=A"!!,C<!!#D!J,hq_!!#<Hz!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	SetVariable svMin,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	SetVariable svMin,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	CheckBox chkAutoMax,userdata(ResizeControlsInfo)=A"!!,GX!!#D+J,hna!!#<8z!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	CheckBox chkAutoMax,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	CheckBox chkAutoMax,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	SetVariable svMax,userdata(ResizeControlsInfo)=A"!!,C4!!#D+J,hq`!!#<Hz!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	SetVariable svMax,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	SetVariable svMax,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	Slider sliderMin,userdata(ResizeControlsInfo)=A"!!,CD!!#D&J,hrK!!#<Hz!!#](Aon\"Qzzzzzzzzzzzzzz!!#o2B4uAezz"
	Slider sliderMin,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Slider sliderMin,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	Slider sliderMax,userdata(ResizeControlsInfo)=A"!!,CD!!#D0J,hrK!!#<Hz!!#](Aon\"Qzzzzzzzzzzzzzz!!#o2B4uAezz"
	Slider sliderMax,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Slider sliderMax,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	GroupBox grpPixel,userdata(ResizeControlsInfo)=A"!!,HD!!#Ck5QF-P!!#@^z!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	GroupBox grpPixel,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	GroupBox grpPixel,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	PopupMenu popColor,userdata(ResizeControlsInfo)=A"!!,HLJ,htF^]6\\l!!#<Xz!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	PopupMenu popColor,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	PopupMenu popColor,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	Button btnZap,userdata(ResizeControlsInfo)=A"!!,HLJ,htL5QF,A!!#<hz!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	Button btnZap,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Button btnZap,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	Button btnCopy,userdata(ResizeControlsInfo)=A"!!,HLJ,htRJ,ho,!!#<hz!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	Button btnCopy,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Button btnCopy,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	Button btnPaste,userdata(ResizeControlsInfo)=A"!!,HLJ,htX^]6\\l!!#<hz!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	Button btnPaste,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Button btnPaste,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
		
	SetWindow kwTopWin,userdata(ResizeControlsInfo)=A"!!*'\"z!!#C%J,htc^]4?7zzzzzzzzzzzzzzzzzzzz"
	SetWindow kwTopWin,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzzzzzzzzzzzzzzz"
	SetWindow kwTopWin,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzzzzzzzzz!!!"
//	SetWindow kwTopWin,userdata(ResizeControlsGuides)="FGT;FGM;FGB;"
//	SetWindow kwTopWin,userdata(ResizeControlsInfoFGT)="NAME:FGT;WIN:HyperspecBrowser;TYPE:User;HORIZONTAL:1;POSITION:30.00;GUIDE1:FT;GUIDE2:;RELPOSITION:30;"
//	SetWindow kwTopWin,userdata(ResizeControlsInfoFGM)="NAME:FGM;WIN:HyperspecBrowser;TYPE:User;HORIZONTAL:1;POSITION:182.00;GUIDE1:FT;GUIDE2:FB;RELPOSITION:0.27;"
//	SetWindow kwTopWin,userdata(ResizeControlsInfoFGB)="NAME:FGB;WIN:HyperspecBrowser;TYPE:User;HORIZONTAL:1;POSITION:525.00;GUIDE1:FB;GUIDE2:;RELPOSITION:-150;"

	// resizing panel hook
	SetWindow HyperspecBrowser hook(ResizeControls)=ResizeControls#ResizeControlsHook
	
	SetWindow HyperspecBrowser userdata(version) = num2str(GetThisVersion())
		
	if (prefs.restart)
		DoUpdate
		MoveWindow/W=HyperspecBrowser left, top, oldright, oldbottom // points
	endif
		
	prefs.restart = 0
	SavePackagePreferences ksPackageName, ksPrefsFileName, 0, prefs
end

static function cpu2point(variable cpu)
	return ScreenResolution > 96 ? cpu  : cpu * 72 / ScreenResolution
end

static function point2cpu(variable point)
	return ScreenResolution > 96 ? point : point * ScreenResolution / 72
end

function InfoMenu()
	string cmd = "(Version: " + num2str(GetThisVersion()/100) + ";"
	cmd += "Visit web page;Email the developer;"
	PopupContextualMenu cmd
	if (V_flag == 2)
		BrowseURL /Z "https://www.wavemetrics.com/node/" + num2str(kProjectID)
	elseif (V_flag == 3)
		Make/T/free/N=5 wt = "mail"
		string newline = "\r\n"
		wt[3] = "subject=" + URLEncode(ksShortTitle + " Package for Igor Pro")
		wt[4] = "Type your message here" + newline + newline
		wt[4] += "PROJECT VERSION:" + num2str(GetThisVersion()/100) + newline
		wt[4] += ReplaceString(";", IgorInfo(3), newline) + newline
		wt[4] = "body=" + URLEncode(wt[4])
		wt[1] = "tony withers"
		wt[2] = "uni-bayreuth de"
		sprintf cmd, "%sto:%s%%%d%s?%s&%s" wt[0], wt[1], 40, wt[2], wt[3], wt[4]
		BrowseURL/Z ReplaceString(" ", cmd, ".")
	endif
end

menu "GraphMarquee"
	MarqueeMenu("Save Average..."), /Q, Hyperspec#SaveAverage()
end

function/S MarqueeMenu(string s)
	GetMarquee/Z
	if (cmpstr(S_marqueeWin, "HyperspecBrowser#G0"))
		return ""
	endif
	return s
end

static function SaveAverage()
	DFREF dfr = getDFR()
	wave w = dfr:wAv
	SaveWaveUI(w, basename=NameOfWave(GetImageWave()) + "_av")
end

// display a spectrum representing average of selected pixels
static function AveragePixelsInRect(STRUCT Rect &pix)
	wave w3D = GetImageWave()
	Duplicate/free/RMD=[pix.left,pix.right][pix.bottom,pix.top][] w3D wROI
	MatrixOP/free w = sumRows(sumCols(wROI)/numCols(wROI))/numRows(wROI)
	
	wave/SDFR=GetDFR() spectrum
	spectrum = w[0][0][p]
	
	string strTextBox = ""
	sprintf strTextBox, "%s[%g,%g][%g,%g]", NameOfWave(w3D), pix.left, pix.right, pix.bottom, pix.top
	TextBox/C/W=HyperspecBrowser#G1/N=text1 strTextBox
	note/K spectrum "source=" + strTextBox
end

static function hookHyperspec(STRUCT WMWinHookStruct &s)
	
	if (s.eventcode == 2) // kill
		PrefsSave()
	endif
	
	if (s.eventcode == 7) // cursor
		strswitch(s.winName)
		 case "HyperspecBrowser#G0" : // cursor A in image plot
		 	if (cmpstr(s.cursorname, "A"))
		 		break
		 	endif
		 	// plot the spectrum for the pixel under the cursor
		 	ResetCrosshairSpectrum()
		 	break
		 case "HyperspecBrowser#G1" : // cursor B in spectrum plot
		 	if (cmpstr(s.cursorname, "B"))
		 		break
		 	endif
		 	// sync setvar with cursor
		 	variable xval = hcsr($s.cursorname, s.winName)
			SetVariable svPlane win=HyperspecBrowser, value=_NUM:xval // this doesn't trigger update event in setvar proc
			ResetPlots(3)
		 	break
		 endswitch
	endif
	
	GetWindow/Z HyperspecBrowser hook(hPixels)
	if (strlen(S_Value))
		return 0
	endif
	
	if (s.eventcode == 5 && cmpstr(s.winname, "HyperspecBrowser#G0") == 0) // mouseup
		STRUCT Rect MQPixels
		if (Marquee2pixels(MQPixels))
			// a marquee is present
			AveragePixelsInRect(MQPixels) // display a spectrum representing average of selected pixels
			DisplayMask(2, pixels=MQPixels)
			ResetPlots(2) // set color as f(z)
		else
			GetWindow/Z HyperspecBrowser hook(hClosedPath)
			if (strlen(S_Value) == 0)
				DisplayMask(0)
			endif
			ResetCrosshairSpectrum() // resets spectrum IF cursor is on graph; updates plot
		endif
	endif
	
	if (s.eventcode == 11)	// Keyboard event

		if (s.keycode == 65)
			SetAxis/W=HyperspecBrowser#G0/A
			STRUCT PackagePrefs prefs
			PrefsLoad(prefs)
			if (prefs.options&1)
				SetAxis/A/R/W=HyperspecBrowser#G1 bottom
			else
				SetAxis/A/W=HyperspecBrowser#G1 bottom
			endif
			SetAxis/A/W=HyperspecBrowser#G1 left
			return 1
		endif
		
		if (s.keycode == 90)
			UndoZap()
			return 1
		endif
		
	endif
	
	return 0
end

// plot the spectrum for the pixel under the cursor
static function ResetCrosshairSpectrum()
	if (strlen(CsrInfo(A, "HyperspecBrowser#G0"))==0)
		ResetPlots(2)
		return 0
	endif
	variable pp = pcsr(A, "HyperspecBrowser#G0")
	variable qq = qcsr(A, "HyperspecBrowser#G0")
	wave w3D = GetImageWave()
	wave/SDFR=GetDFR() spectrum
	spectrum = w3D[pp][qq][p]
	string strTextBox = ""
	if (hcsr(A, "HyperspecBrowser#G0")!=pp || vcsr(A, "HyperspecBrowser#G0")!=qq)
		sprintf strTextBox, "%s[%g][%g] x=%g y=%g", NameOfWave(w3D), pp, qq, hcsr(A, "HyperspecBrowser#G0"), vcsr(A, "HyperspecBrowser#G0")
	else
		sprintf strTextBox, "%s[%g][%g]", NameOfWave(w3D), pp, qq
	endif
	TextBox/C/W=HyperspecBrowser#G1/N=text1 strTextBox
	SetWindow HyperspecBrowser#G0 userdata(hpixel)=num2str(pp)
	SetWindow HyperspecBrowser#G0 userdata(vpixel)=num2str(qq)
	
	string strNote = ""
	sprintf strNote "source=%s[%d][%d]", NameOfWave(w3D), pp, qq
	note/K spectrum strNote
	
	ResetPlots(2)
	return 1
end

static function Marquee2pixels(STRUCT Rect &MQPixels)
	GetMarquee/Z/W=HyperspecBrowser#G0 left, bottom
	if (V_flag)
		wave w3D = GetImageWave()
		int p1, p2, q1, q2, pmin, pmax, qmin, qmax
		q1 = scaleToIndex(w3D, v_top, 1)
		q2 = scaleToIndex(w3D, v_bottom, 1)
		p1 = scaleToIndex(w3D, v_left, 0)
		p2 = scaleToIndex(w3D, v_right, 0)
		pmin = 0
		qmin = 0
		pmax = DimSize(w3D, 0) - 1
		qmax = DimSize(w3D, 1) - 1
		MQPixels.left = max(pmin, min(p1, p2))
		MQPixels.right = min(pmax, max(p1, p2))
		MQPixels.bottom = max(min(q1, q2), qmin)
		MQPixels.top = min(max(q1, q2), qmax)
	endif
	return v_flag
end

static function hookPixelBrush(STRUCT WMWinHookStruct &s)
	
	if (cmpstr(s.winname, "HyperspecBrowser#G0"))
		return 0
	endif
	
	if (s.eventMod & 16) // allow contextual menu
		return 0
	endif
	
	// need to fix this
	if (s.eventMod & 13) // cmd/ctrl
		s.doSetCursor = 1
//		s.cursorCode = 20, 15, 18
		s.cursorCode = (s.eventMod & 4) ? 19 : 18
	endif
	
	if (s.eventCode==3 && !(s.eventMod&12) )
		PixelEdit(0)
	endif
		
	if (! (s.eventCode==3 && (s.eventMod&12)) ) // mousedown, alt/ctrl/shift-ctrl
		return 0
	endif
	
	wave w_img = GetImageWave()
	int pmax = DimSize(w_img, 0)
	int qmax = DimSize(w_img, 1)
	
	DFREF dfr = GetDFR()
	wave/SDFR=dfr wMask, wDisplayMask
	
	variable v_X, v_Y, p_X, p_Y

	int ctrl = (s.eventMod & 8) > 0
	int opt = (s.eventMod & 4) > 0
	int keys, shift
	keys = ctrl + 2*opt

	int mousebutton = 1
	int oldLeft = s.mouseLoc.h, oldTop = s.mouseLoc.v
	int horiz = 0, vert = 0
	int firstLoop = 1
	for (;mousebutton && (keys&3);)
		keys = GetKeyState(0)
		GetMouse/W=HyperspecBrowser //$s.WinName
		mousebutton = V_flag & 1
		
		if (keys & 4) // shift
			if ((horiz + vert) == 0)
				if (abs(v_left-oldLeft) > abs(v_top-oldTop))
					vert = 1
				elseif (abs(v_top-oldTop) > abs(v_left-oldLeft))
					horiz = 1
				endif
			endif
			if (vert)
				v_top = oldTop
			elseif (horiz)
				v_left = oldLeft
			endif
		else
			horiz = 0
			vert = 0
		endif
				
		if (v_left==oldLeft && v_top==oldTop && (!firstLoop))
			continue
		endif
		firstLoop = 0
		
		oldLeft = v_left
		oldTop = v_top
		
		ctrl = keys & 1
		opt = keys & 2
		if (!(mousebutton && (ctrl||opt)))
			continue
		endif
		
		variable hh = AxisValFromPixel("HyperspecBrowser#G0", "bottom", v_left)
		variable vv = AxisValFromPixel("HyperspecBrowser#G0", "left", v_top)
		
		int pp = scaleToIndex(w_img, hh, 0)
		int qq = scaleToIndex(w_img, vv, 1)
		
		if (pp<0 || qq<0 || pp>=pmax || qq>=qmax)
			continue
		endif
	
		wMask[pp][qq] = ctrl
		wDisplayMask[pp][qq][3] = 40000 * !ctrl
		DoUpdate

	endfor
	AveragePixelsROI(wMask)
	
	return 0
end

static function PopupProc(STRUCT WMPopupAction &s) : PopupMenuControl
	
	if (s.eventCode != 2)
		return 0
	endif
	
	if (cmpstr(s.ctrlname, "popCT") && cmpstr(s.ctrlname, "popColor"))
		return 0
	endif
	
	ResetPlots(3)
	
	return 0
end

static function SetVariableProc(STRUCT WMSetVariableAction &s) : SetVariableControl

	if (s.eventCode<1 || s.eventCode>8)
		return 0
	endif
	
	variable vMin, vMax, vPos
	
	strswitch (s.ctrlName)
		case "svPlane":
			GetAxis/W=HyperspecBrowser#G1/Q bottom
			Cursor/W=HyperspecBrowser#G1/F/P B spectrum  (s.dval-V_min)/(V_max-V_min), 0.5
		case "svMin":
		case "svMax": // plane, min or max
			ResetPlots(3)
			break
		case "svDatacube" :
			if (s.eventcode == 7) // begin edit

					CreateBrowser/M Prompt="Select a 3D wave that represents a hyperspectral datacube"
					ModifyBrowser/M showWaves=1, showVars=0, showStrs=0
					ModifyBrowser/M ShowInfo=0, showPlot=0
					ModifyBrowser/M clearSelection
					if (strlen(s.sval))
						ModifyBrowser/M selectList=s.sval
					endif
					ModifyBrowser/M showModalBrowser
					if (V_Flag == 0)
						// use ControlUpdate to deselect the setvar
						ControlUpdate/W=$s.win $s.ctrlname
						return 0
					endif
					if (ItemsInList(S_BrowserList) == 1)
						wave/Z wSel = $StringFromList(0, S_BrowserList)
						MakeHyperspecBrowser(wSel)
						ControlUpdate/W=$s.win $s.ctrlname
					endif
			elseif (s.eventcode == 8) // endedit
				wave/Z w = $s.sval
				if (Valid3DWave(w) && (!WaveRefsEqual(w, GetImageWave())))
					MakeHyperspecBrowser(w)
				else
					SetVariable $s.ctrlName, win=$s.win, value=_STR:GetUserData(s.win, s.ctrlname, "")
				endif
			
			endif
			break
	endswitch
		
	return 0
end

static function ResetSliders([variable limitMin, variable limitMax, variable vMin, variable vMax])
		
	variable vPos
	
	if (ParamIsDefault(vMin))
		ControlInfo/W=HyperspecBrowser svMin
		vMin = V_Value
	endif
	
	if (ParamIsDefault(vMax))
		ControlInfo/W=HyperspecBrowser svMax
		vMax = V_Value
	endif
	
	if (ParamIsDefault(limitMin) || ParamIsDefault(limitMax))
		ControlInfo/W=HyperspecBrowser sliderMin
		vPos = strsearch(s_recreation, "limits=", 0)
		variable v1, v2
		sscanf s_recreation[vPos,Inf], "limits={%g,%g,0}", v1, v2
		limitMin = ParamIsDefault(limitMin) ? v1 : limitMin
		limitMax = ParamIsDefault(limitMax) ? v2 : limitMax
	endif
	
	Slider sliderMin win=HyperspecBrowser, value=limit(vMin, limitMin, limitMax), limits={limitMin,limitMax,0}
	Slider sliderMax win=HyperspecBrowser, value=limit(vMax, limitMin, limitMax), limits={limitMin,limitMax,0}
end


static function SliderProc(STRUCT WMSliderAction &s) : SliderControl

	if (s.eventcode>0 && s.eventcode&1)
		
		if (cmpstr(s.ctrlname, "sliderMin") == 0)
			SetVariable svMin win=HyperspecBrowser, value=_NUM:s.curval
		elseif (cmpstr(s.ctrlname, "sliderMax") == 0)
			SetVariable svMax win=HyperspecBrowser, value=_NUM:s.curval
		endif
		ResetPlots(3)
	endif
	
	if (s.eventmod == 4)
		
	endif
	
	return 0
end


// set plot plane and color table based on control values
static function ResetPlots(int whichone) // bit 0 = image plot, bit 1 = spectrum
	
	int autoMin, autoMax, plane
	variable wavenumber
	variable vMin, vMax, planeMin, planeMax
		
	ControlInfo/W=HyperspecBrowser chkAutoMin
	autoMin = V_Value
	ControlInfo/W=HyperspecBrowser chkAutoMax
	autoMax = V_Value
		
	string strImage = StringFromList(0, ImageNameList("HyperspecBrowser#G0", ";"))
	wave w3D = ImageNameToWaveRef("HyperspecBrowser#G0", strImage)
	
	ControlInfo/W=HyperspecBrowser svPlane
	
	plane = scaleToIndex(w3D, v_value, 2)
	wavenumber = IndexToScale(w3D, plane, 2)
	
	WaveStats/Q/M=1/RMD=[][][plane] w3D
	planeMin = v_min
	planeMax = v_max
		
	if (autoMin)
		vMin= planeMin
		SetVariable svMin win=HyperspecBrowser, value=_NUM:vMin
	else
		ControlInfo/W=HyperspecBrowser svMin
		vMin = v_value
	endif
	if (autoMax)
		vMax = planeMax
		SetVariable svMax win=HyperspecBrowser, value=_NUM:vMax
	else
		ControlInfo/W=HyperspecBrowser svMax
		vMax = v_value
	endif
	
	ControlInfo/W=HyperspecBrowser popCT
	string Table = s_value
	
	if (whichone & 1)
		ModifyImage/W=HyperspecBrowser#G0 ''#0 plane=plane, ctab={vMin,vMax,$Table,0}
		ResetPixelTextbox()

		ControlInfo/W=HyperspecBrowser PopColor
		variable red, green, blue
		sscanf s_value, "(%d,%d,%d)", red, green, blue
		Cursor/M/W=HyperspecBrowser#G0/C=(red, green, blue) A

		if (WaveExists(TraceNameToWaveRef("HyperspecBrowser#G0", "ROI_nodes")))
			ModifyGraph/Z rgb(ROI_nodes)=(red, green, blue)
			wave/Z/SDFR=GetDFR() ROI_x, ROI_y
			Wave2Poly(ROI_x, ROI_y, "bottom", "left", layer="ProgFront", name="ROI")
		endif

		string strTextBox = ""
		sprintf strTextBox, "%s[][][%g]", NameOfWave(w3D), plane
		TextBox/C/W=HyperspecBrowser#G0/N=text0 strTextBox
	endif
	
	if (whichone & 2)
		wave spectrum = TraceNameToWaveRef("HyperspecBrowser#G1", "spectrum")
		ModifyGraph/W=HyperspecBrowser#G1 zColor(spectrum)={spectrum,vMin,vMax,$Table,0}
		ResetSliders(limitMin=planeMin, limitMax=planeMax, vMin=vMin, vMax=vMax)
	endif
		
	SetVariable svPlane, win=HyperspecBrowser, value=_NUM:wavenumber
		
end

// returns position (proportion of axis) at axis value = val
static function AxisValToPosition(variable val)
	GetAxis/W=HyperspecBrowser#G1/Q bottom
	return (val-V_min)/(V_max-V_min)
end

static function CheckBoxProc(STRUCT WMCheckBoxAction &s) : CheckboxControl
	
	if (s.eventCode != 2)
		return 0
	endif
	
	strswitch (s.ctrlName)
		case "chkAutoMin":
			SetVariable svMin  win=HyperspecBrowser, disable=2*s.checked
			Slider sliderMin   win=HyperspecBrowser, disable=2*s.checked
			if (s.checked)
				ResetPlots(3)
			endif
			break
		case "chkAutoMax":
			SetVariable svMax  win=HyperspecBrowser, disable=2*s.checked
			Slider sliderMax   win=HyperspecBrowser, disable=2*s.checked
			if (s.checked)
				ResetPlots(3)
			endif
			break
		case "chkLog" :
			
			logcolors(s.checked)
//			string strImage = StringFromList(0,ImageNameList("HyperspecBrowser#G0", ";"))
//			ModifyImage/W=HyperspecBrowser#G0 $strImage log=s.checked
//			ColorScale/W=HyperspecBrowser#G0/C/N=scale log=s.checked
			break
	endswitch
		
	return 0
end


static function logcolors(int isLog)
	string strImage = StringFromList(0,ImageNameList("HyperspecBrowser#G0", ";"))
	ModifyImage/W=HyperspecBrowser#G0 $strImage log=isLog
	ColorScale/W=HyperspecBrowser#G0/C/N=scale log=isLog
end

static function ButtonProc(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode != 2)
		return 0
	endif
	
	strswitch (s.ctrlName)
		case "btnZap":
			ZapPixel()
			break
		case "btnCopy":
			CopyPastePixel(0)
			break
		case "btnPaste":
			CopyPastePixel(1)
			break
		case "btnInfo":
			InfoMenu()
			break
		case "btnSettings":
			if (!WinType("HyperspecPrefsPanel"))
				MakePrefsPanel()
			else
				PrefsClosePanel(0)
			endif
			
			break
		case "btnSave":
			PrefsClosePanel(1)
			break
		case "btnCancel":
			PrefsClosePanel(0)
			break
	endswitch
	
	return 0
end

static function CopyPastePixel(int paste)
	DFREF dfr = GetDFR()
	wave/Z pixelcopy = dfr:pixelcopy
	wave/Z w3D = GetImageWave()
	variable pp = pcsr(A, "HyperspecBrowser#G0")
	variable qq = qcsr(A, "HyperspecBrowser#G0")
	
	if (paste)
		if (!WaveExists(pixelcopy) || numpnts(pixelcopy)!=DimSize(w3D, 2))
			return 0
		endif
		
		Duplicate/O/RMD=[pp][qq][] w3D dfr:backup/wave=bup
		string strNote = ""
		sprintf strNote, "%s;%g;%g;", NameOfWave(w3D), pp, qq
		note/K bup strNote
		
		w3D[pp][qq][] = pixelcopy[r]
	else
		Duplicate/O/RMD=[pp][qq][] w3D dfr:pixelcopy
	endif
end

static function ZapPixel()
	
	// better to check wavenote?
	string str = ""
	#ifdef WINDOWS
	str = "Note: pixel-zapping is for ratioed data!\r\rDo you want to continue?\r\rCtrl-Z to undo."
	#else
	str = "Note: pixel-zapping is for ratioed data!\r\rDo you want to continue?\r\rCmd-Z to undo."
	#endif
	DoAlert 1, str
	if (v_flag == 2)
		return 0
	endif
	
	variable pp = pcsr(A, "HyperspecBrowser#G0")
	variable qq = qcsr(A, "HyperspecBrowser#G0")
	
	

	wave/Z w3D = GetImageWave()
	
	if (numtype(pp) || numtype(qq) || !WaveExists(w3D))
		return 0
	endif
	
	variable i, j
	Make/free/N=(DimSize(w3D, 2)) total
	variable count
		
	variable newX, newY
	variable xMax = DimSize(w3D, 0)
	variable yMax = DimSize(w3D, 1)
	
	for (i=-1;i<2;i++)
		for (j=-1;j<2;j++)
			newX = pp + i
			newY = qq + j
			if (newX>=0 && newY>=0 && newX<xMax && newY<Ymax && (i || j))
				count += 1
				total += w3D[newX][newY][p]
			endif
		endfor
	endfor
	total /= count
	DFREF dfr = GetDFR()
	Duplicate/O total dfr:backup/wave=bup
	bup = w3D[pp][qq][p]
	string strNote = ""
	sprintf strNote, "%s;%g;%g;", NameOfWave(w3D), pp, qq
	note/K bup strNote
	w3D[pp][qq][] = total[r]
	ResetCrosshairSpectrum()
end

static function/wave GetImageWave()
	string strImage = StringFromList(0, ImageNameList("HyperspecBrowser#G0", ";"))
	wave/Z w3D = ImageNameToWaveRef("HyperspecBrowser#G0", strImage)
	return w3D
end

static function UndoZap()
	wave/Z w3D = GetImageWave()
	wave/SDFR=GetDFR() backup
	if (!WaveExists(backup) || !WaveExists(w3D))
		return 0
	endif
	
	string strNote = note(backup)
	if (cmpstr(NameOfWave(w3D), StringFromList(0, strNote)))
		return 0
	endif
	
	variable pp = str2num(StringFromList(1, strNote))
	variable qq = str2num(StringFromList(2, strNote))
	
	w3D[pp][qq][] = backup[r]
	ResetCrosshairSpectrum()
end


// *** save wave ***

static function/WAVE SaveWaveUI(wave w, [string basename, int overwrite])
	
	overwrite = ParamIsDefault(overwrite) ? 1 : overwrite // default is to allow user to overwrite a wave
	if (ParamIsDefault(basename))
		basename = NameOfWave(w)
		if (cmpstr(basename, "_free_") == 0)
			basename = "wave"
		endif
	endif
	KillWaves/Z wwSavedWaves
	DoWindow/K SaveWaveDialog
	GetMouse
	variable left, top, right, bottom
	left = pixel2cpu(V_left) - 50
	top = pixel2cpu(V_top) - 50
	right = left + 220
	bottom = top + 150
	NewPanel/K=1/W=(left,top,right,bottom)/N=SaveWaveDialog as "Save Wave"
	SetVariable svName,pos={25,20},size={157,14},title="Wave Name", userdata(base) = basename
	SetVariable svName,value=_STR:UniqueName(basename, 1, 0), fsize=12, Proc=Hyperspec#SaveWaveSetVars
	PopupMenu popFolder,pos={25,50},size={157,14},title="Data Folder", fsize=12
	PopupMenu popFolder,value=#"GetDataFolder(0)", userdata(folder)=GetDataFolder(1), Proc=Hyperspec#SaveWavePopups
	Button btnCancel,pos={40,100},size={55,20},title="Cancel", Proc=Hyperspec#SaveWaveButtons
	Button btnDoIt,pos={130,100},size={50,20},title="Do It", Proc=Hyperspec#SaveWaveButtons, userdata=num2str(overwrite)

	SetVariable svName,userdata(ResizeControlsInfo)=A"!!,C,!!#<X!!#A,!!#<Hz!!#](Aon\"Qzzzzzzzzzzzzzz!!#o2B4uAezz"
	SetVariable svName,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#u:Du]k<zzzzzzzzzzz"
	SetVariable svName,userdata(ResizeControlsInfo)+=A"zzz!!#u:Du]k<zzzzzzzzzzzzzz!!!"
	PopupMenu popFolder,userdata(ResizeControlsInfo)=A"!!,C,!!#>V!!#@`!!#<Xz!!#](Aon\"Qzzzzzzzzzzzzzz!!#o2B4uAezz"
	PopupMenu popFolder,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#u:Du]k<zzzzzzzzzzz"
	PopupMenu popFolder,userdata(ResizeControlsInfo)+=A"zzz!!#u:Du]k<zzzzzzzzzzzzzz!!!"
	Button btnCancel,userdata(ResizeControlsInfo)=A"!!,D/!!#@,!!#>j!!#<Xz!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	Button btnCancel,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Button btnCancel,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	Button btnDoIt,userdata(ResizeControlsInfo)=A"!!,Fg!!#@,!!#>V!!#<Xz!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	Button btnDoIt,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Button btnDoIt,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	SetWindow kwTopWin,userdata(ResizeControlsInfo)=A"!!*'\"z!!#Ak!!#A%zzzzzzzzzzzzzzzzzzzzz"
	SetWindow kwTopWin,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzzzzzzzzzzzzzzz"
	SetWindow kwTopWin,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzzzzzzzzz!!!"
	
	// resizing panel hook
	SetWindow SaveWaveDialog hook(ResizeControls)=ResizeControls#ResizeControlsHook

	PauseForUser SaveWaveDialog
	wave/Z/WAVE wwSavedWaves
	if (WaveExists(wwSavedWaves))
		wave/Z saved = wwSavedWaves[0]
		string strPathToNewWave = GetWavesDataFolder(saved, 2)
		KillWaves/Z saved
		Duplicate/O w $strPathToNewWave /wave=saved
		KillWaves wwSavedWaves
		return saved
	endif
	return $""
end

static function SaveWavePopups(STRUCT WMPopupAction &s) : PopupMenuControl
	
	if (s.eventCode != 3)
		return 0
	endif

	string strFolder = StringFromList(0, SelectFolder(s.popStr))
	if (strlen(strFolder))
		PopupMenu popFolder, win=$s.win, userdata(folder)=strFolder
		PopupMenu popFolder, win=$s.win, value=#"Hyperspec#GetSelectedFolder(0)"
	endif
	
	CheckNameSetvar()

	return 0
end

static function SaveWaveButtons(STRUCT WMButtonAction &s) : ButtonControl

	if (s.eventcode != 2)
		return 0
	endif
		
	strswitch (s.ctrlName)
		case "btnDoIt":
			int overwrite = str2num(s.userdata)
			ControlInfo/W=$s.win svName
			if (!overwrite && exists(GetSelectedFolder(1)+PossiblyQuoteName(s_value)))
				DoAlert 0, s_value + " exists."
				return 0
			endif
			KillWaves/Z $GetSelectedFolder(1)+PossiblyQuoteName(s_value)
			wave/Z w = $GetSelectedFolder(1)+PossiblyQuoteName(s_value)
			if (WaveExists(w))
				DoAlert 0, s_value + " exists and is in use, cannot overwrite."
				return 0
			endif
			Make/O/N=0 $GetSelectedFolder(1)+PossiblyQuoteName(s_value) /wave=w
			Make/O/WAVE wwSavedWaves = {w}
		case "btnCancel" : // Cancel or Do It
			DoWindow/K $s.win
			break
	endswitch
	return 0
end

static function SaveWaveSetVars(STRUCT WMSetVariableAction &s)
	
	if (s.eventCode < 0) // excludes Igor8 eventcodes -2 and -3
		return 0
	endif
	
	if (cmpstr(s.ctrlName, "svName") == 0)
		CheckNameSetvar()
	endif
	return 0
end

static function CheckNameSetvar()
	ControlInfo/W=SaveWaveDialog svName
	if (strlen(S_Value) == 0)
		S_Value = UniqueName(GetUserData("SaveWaveDialog", "svName", "base"), 1, 0)
		SetVariable svName win=SaveWaveDialog, value=_STR:S_Value
	endif
	if (exists(GetSelectedFolder(1) + S_Value))
		SetVariable svName win=SaveWaveDialog, valueBackColor=(0xFFFF,0,0)
	else
		SetVariable svName win=SaveWaveDialog, valueBackColor=0
	endif
end

static function/T GetSelectedFolder(int fullpath)
	string strFolder = GetUserData("SaveWaveDialog", "popFolder", "folder")
	return SelectString(fullpath, ParseFilePath(0, strFolder, ":", 1, 0), strFolder)
end

static function/T SelectFolder(string selection)
	CreateBrowser/M Prompt="Select Data Folder"
	ModifyBrowser/M showWaves=0, showVars=0, showStrs=0
	ModifyBrowser/M ShowInfo=0, showPlot=0
	ModifyBrowser/M select=selection
	ModifyBrowser/M showModalBrowser
	if (V_Flag == 0)
		return "" // User cancelled
	endif
	return S_BrowserList
end

static function pixel2cpu(variable pixel, [string win])
	variable expansion = ParamIsDefault(win) ? 1 : PanelResolution(win)/PanelResolution("")
	return (ScreenResolution > 96 ? pixel * 72 / ScreenResolution : pixel) / expansion
end

static function/S PolyMenu(string str)
	
	DFREF dfr = GetDFR()
	wave/Z polyX = dfr:ROI_x
	wave/Z polyY = dfr:ROI_y
	
	if (WaveExists(polyY)==0 || numpnts(polyY)<3)
		return ""
	endif
	
	GetLastUserMenuInfo
	if (cmpstr(S_graphName, "HyperspecBrowser#G0"))
		return ""
	endif
	
	CheckDisplayed/W=HyperspecBrowser#G0 polyY
	if (!V_flag)
		return ""
	endif
	
	if (abs(polyy[numpnts(polyy)-1]-polyy[0]) > 1e-5)
		return ""
	elseif (abs(polyx[numpnts(polyx)-1]-polyx[0]) > 1e-5)
		return ""
	endif
	
	Make/free wx={AxisValFromPixel(S_graphName, "bottom", V_mouseX)}
	Make/free wy={AxisValFromPixel(S_graphName, "left", V_mouseY)}
	FindPointsInPoly wx, wy, polyx, polyy
	wave W_inPoly
	int inPoly = W_inPoly[0]
	KillWaves/Z W_inPoly
	return SelectString(inPoly, "", str)
end

static function inPoly(variable x, variable y)
	DFREF dfr = GetDFR()
	wave wx = dfr:ROI_x
	wave wy = dfr:ROI_y
	Make/free wxp = {x}
	Make/free wyp = {y}
	
	FindPointsInPoly wxp, wyp, wx, wy
	wave W_inPoly
	int inPoly = W_inPoly[0]
	KillWaves/Z W_inPoly
	return inPoly
end

menu "TracePopup", dynamic
	"-"
	Hyperspec#InGraphMenu("Save Spectrum...", "HyperspecBrowser#G1"), /Q, Hyperspec#SaveSpectrum()
	Hyperspec#HookActiveMenu("Draw ROI", "hClosedPath", "HyperspecBrowser#G0", 0), /Q, Hyperspec#StartClosedPath("HyperspecBrowser#G0", "ProgFront", "left", "bottom", "ROI", 1)
	Hyperspec#HookActiveMenu("Stop ROI Drawing", "hClosedPath", "HyperspecBrowser#G0", 1), /Q, Hyperspec#StopClosedPath("")
	Hyperspec#HookActiveMenu("Select Pixels", "hPixels", "HyperspecBrowser#G0", 0), /Q, Hyperspec#PixelEdit(1)
	Hyperspec#HookActiveMenu("Stop Select Pixels", "hPixels", "HyperspecBrowser#G0", 1), /Q, Hyperspec#PixelEdit(0)

	Hyperspec#InGraphMenu("Draw Shapes", "HyperspecBrowser#G0"), /Q, Hyperspec#StartDrawing()
	Hyperspec#InGraphMenu("ROI from Shapes", "HyperspecBrowser#G0"), /Q, Hyperspec#AveragePixelsROI($"")
	Hyperspec#InGraphMenu("Clear Shapes", "HyperspecBrowser#G0"), /Q, Hyperspec#StopClosedPath(""); SetDrawLayer/W=HyperspecBrowser#G0/K ProgFront //;
end

menu "GraphPopup", dynamic
	"-"
	Hyperspec#InGraphMenu("Save Spectrum...", "HyperspecBrowser#G1"), /Q, Hyperspec#SaveSpectrum()
end

static function StartDrawing()
	ControlInfo/W=HyperspecBrowser PopColor
	variable red, green, blue
	sscanf s_value, "(%d,%d,%d)", red, green, blue
	
	ShowTools/A/W=HyperspecBrowser arrow
	SetDrawLayer/W=HyperspecBrowser#G0 ProgFront
	SetDrawEnv/W=HyperspecBrowser#G0 linefgc=(red, green, blue), fillpat=0, xcoord=bottom, ycoord=left	, Save
end

static function/S InGraphMenu(string str, string strGraph)
	GetLastUserMenuInfo
	if (cmpstr(S_graphName, strGraph))
		return ""
	endif
	return str
end

//function/S ShapesMenu(string str)
////	GetLastUserMenuInfo
////	if (cmpstr(S_graphName, strGraph))
////		return ""
////	endif
//	return str
//end

static function/S HookActiveMenu(string str, string hook, string win, int active)
	GetLastUserMenuInfo
	if (cmpstr(S_graphName, win))
		return ""
	endif
	GetWindow/Z $parent(S_graphName) hook($hook)
	return SelectString((strlen(s_value)>0)==active, "", str)
end

static function SaveSpectrum()
	wave w = TraceNameToWaveRef("HyperspecBrowser#G1", "spectrum")
	SaveWaveUI(w)
end

static function ShowCrosshair(int show)
	if (show && strlen(CsrInfo(A, "HyperspecBrowser#G0"))==0)
		wave w3D = GetImageWave()
		variable xpnt = str2num(GetUserData("HyperspecBrowser#G0", "", "hpixel"))
		variable ypnt = str2num(GetUserData("HyperspecBrowser#G0", "", "vpixel"))
		xpnt = numtype(xpnt) ? DimSize(w3D, 0)/2 : xpnt
		ypnt = numtype(ypnt) ? DimSize(w3D, 1)/2 : ypnt
		ControlInfo/W=HyperspecBrowser PopColor
		variable red, green, blue
		sscanf s_value, "(%d,%d,%d)", red, green, blue
		Cursor/W=HyperspecBrowser#G0/P/I/H=1/C=(red, green, blue)/N=1 A $NameOfWave(w3D) xpnt, ypnt
	elseif (!show)
		Cursor/W=HyperspecBrowser#G0/K A
	endif
end

// show = 1: mask points outside of poly drawing object created by closed path hook
// show = 2: mask points outside of marquee
// show = 3: mask all points
// show = 4: sync mask with M_ROIMask created by ImageGenerateROIMask
// show = 0: clear mask
static function DisplayMask(variable show, [STRUCT Rect &Pixels])
	wave/SDFR=GetDFR() wDisplayMask
	switch (show)
		case 0 :
			wDisplayMask[][][3] = 0
			break
		case 1 :
			wDisplayMask[][][3] = 40000 * !inPoly(p, q)
			break
		case 2 :
			wDisplayMask[][][3] = 40000 * !((p+0.5)==limit(p+0.5, pixels.left, pixels.right) && (q+0.5)==limit((q+0.5), pixels.bottom, pixels.top))
			wDisplayMask[][][3] = 40000 * !((p)==limit(p, pixels.left, pixels.right) && (q)==limit((q), pixels.bottom, pixels.top))
			break
		case 3 :
			wDisplayMask[][][3] = 40000
			break
		case 4 :
			wDisplayMask[][][3] = 40000
			wave/Z M_ROIMask
			if (WaveExists(M_ROIMask))
				wDisplayMask[][][3] = 40000 * !M_ROIMask[p][q]
			endif
			break
		default:
			wDisplayMask[][][3] = 0
	endswitch
	ShowCrosshair(!show)
end

// if wMask is null, an image mask is created from drawing layer object(s)
static function AveragePixelsROI(wave/Z wMask)
		
	if (!WaveExists(wMask))
		KillWaves/Z M_ROIMask
		ImageGenerateROIMask/W=HyperspecBrowser#G0 $StringFromList(0,ImageNameList("HyperspecBrowser#G0", ";"))
		if (!v_flag)
			DoAlert 0, "No drawing objects to analyse"
			return 0
		endif
		wave wMask = M_ROIMask
		DisplayMask(4)
	endif
	
	wave w3D = GetImageWave()
	Duplicate/free w3D wROI
	wROI = wMask[p][q] ? wROI : 0
	
	MatrixOP/free w = sumRows(sumCols(wROI))
	variable numSpectra = sum(wMask)
	
	wave/SDFR=GetDFR() spectrum
	spectrum = w[0][0][p] / numSpectra
	
	string strTextBox = ""
	sprintf strTextBox, "Masked Region"
	TextBox/C/W=HyperspecBrowser#G1/N=text1 strTextBox
	
	note/K spectrum "source=" + NameOfWave(w3D) + " masked region"
	
	ResetPlots(2)
	
	KillWaves/Z M_ROIMask
end

static function PixelEdit(int start)
	if (start)
		StopClosedPath("HyperspecBrowser#G0")
		wave/SDFR=GetDFR() wMask
		wMask = 0
		DisplayMask(3)
		SetWindow HyperspecBrowser hook(hPixels)=Hyperspec#hookPixelBrush
		ResetPixelTextbox()
		AveragePixelsROI(wMask)
	else
		SetWindow HyperspecBrowser hook(hPixels)=$""
		ResetPixelTextbox()
		DisplayMask(0)
		ResetCrosshairSpectrum()
	endif
end

static function ResetPixelTextbox()
	GetWindow/Z HyperspecBrowser hook(hPixels)
	if (strlen(s_value))
		ControlInfo/W=HyperspecBrowser PopColor
		variable red, green, blue
		sscanf s_value, "(%d,%d,%d)", red, green, blue
		string strTitle = ""
		#ifdef WINDOWS
		sprintf strTitle, "\\K(%d,%d,%d)Ctrl to select, Alt to deselect", red, green, blue
		#else
		sprintf strTitle, "\\K(%d,%d,%d)Cmd to select, Option to deselect", red, green, blue
		#endif
		TextBox/W=HyperspecBrowser#G0/C/N=textPixels/F=0/A=LT/B=1 strTitle
	else
		TextBox/W=HyperspecBrowser#G0/K/N=textPixels
	endif
end

// *** draw shape ***

static function StopClosedPath(string strGraph)
	if (strlen(strGraph)==0)
		GetLastUserMenuInfo
		strGraph = S_graphName
	endif
	SetWindow $parent(strGraph) hook(hClosedPath)=$""
	wave/Z/SDFR=getDFR()/WAVE refs
	if (WaveExists(refs))
		RemoveFromGraph/Z/W=$strGraph $NameOfWave(refs[0]), $NameOfWave(refs[2])
	endif
	DrawAction/L=ProgFront getgroup=ROI, delete
	DisplayMask(0)
	
	ResetCrosshairSpectrum()
end
	
static function Wave2Poly(wave/Z wx, wave/Z wy, string hAxis, string vAxis, [string layer, string name])
	
	layer = SelectString(ParamIsDefault(layer), layer, "ProgFront")
	name = SelectString(ParamIsDefault(name), name, "")
	
	if (DimSize(wy, 0)<2)
		return 0
	endif
	
	name = SelectString(strlen(name)==0, name, NameOfWave(wy) + "Poly")
	ControlInfo/W=HyperspecBrowser PopColor
	variable red, green, blue
	sscanf s_value, "(%d,%d,%d)", red, green, blue
	
	DrawAction/L=$layer getgroup=$name, delete
	SetDrawLayer $layer
	SetDrawEnv gstart, gname=$name
	SetDrawEnv xcoord=$hAxis, ycoord=$vAxis
	SetDrawEnv fillpat= 0, linefgc= (red, green, blue)
	DrawPoly wx[0], wy[0], 1, 1, wx, wy
	SetDrawEnv gstop, gname=$name

	WaveStats/Q/M=1 wy
	return ((V_numNaNs + V_numInfs) == 0)
end


// shape = 1 to create a drawing object.
static function StartClosedPath(string Graph, string layer, string vAxis, string hAxis, string name, int shape)
	
	PixelEdit(0)
	
	DFREF dfr = GetDFR()
	
	Make/O/N=(0,2) dfr:$(name+"_nodes")/wave=nodes
	Make/O/N=(0) dfr:$(name+"_x")/wave=wx
	Make/O/N=(0) dfr:$(name+"_y")/wave=wy
	note wy "layer="+layer+";name="+name+";wx="+GetWavesDataFolder(wx, 2)+";nodes="+GetWavesDataFolder(nodes, 2)+";"
		
	ControlInfo/W=HyperspecBrowser PopColor
	variable red, green, blue
	sscanf s_value, "(%d,%d,%d)", red, green, blue
			
	RemoveFromGraph/Z $NameOfWave(wy)
	AppendToGraph/W=HyperspecBrowser#G0 wy vs wx
	ModifyGraph/Z mode($NameOfWave(wy))=0
	
	string strNodes = NameOfWave(nodes)
	RemoveFromGraph/Z $strNodes // keep trace on top
	AppendToGraph/W=HyperspecBrowser#G0 nodes[][1] vs nodes[][0]
	
	ModifyGraph mode($strNodes)=3, marker($strNodes)=8, rgb($strNodes)=(red, green, blue)
	if (shape) // make spline transparent (not hidden, because we want to be able to right-click and edit nodes)
		ModifyGraph rgb($NameOfWave(wy))=(65535,0,0,0)
	endif
	
	Make/O/WAVE/N=3 dfr:refs/wave=refs
	refs[0] = nodes
	refs[1] = wx
	refs[2] = wy
	
	string/G dfr:win = Graph
	string/G dfr:layer = layer
	variable/G dfr:status = (shape != 0)
	variable/G dfr:pointNum = NaN
	variable/G dfr:mode = 0
	variable/G dfr:success = 1
	
	// clear any other active closed path hook
	wave/T wWins = ListToTextWave(WinList("!"+Graph,";","WIN:65"), ";")
	for (string win : wWins)
		SetWindow $win hook(hClosedPath)=$""
	endfor
	DisplayMask(0)
	
	DoWindow/F $parent(Graph)
	SetActiveSubwindow $Graph
	
	SetWindow $parent(Graph) hook(hClosedPath)=Hyperspec#hookClosedPath
	
	ShowCrosshair(0)
	wave/SDFR=GetDFR() wMask
	wMask = 0
	AveragePixelsROI(wMask)
end

// corrects for aspect ratio of graph area and relative scale of axes.
static function AspectRatio(string strGraph, string hAxis, string vAxis)
	GetAxis/W=$strGraph/Q $hAxis
	variable horizontalAx = V_max - V_min
	
	string strInfo = AxisInfo(strGraph, hAxis)
	if( NumberByKey("log(x)",strInfo,"="))
		horizontalAx = log(V_max) - log(V_min)
	endif
	
	GetAxis/W=$strGraph/Q $vAxis
	variable verticalAx = V_max - V_min
		
	strInfo = AxisInfo(strGraph, vAxis)
	if( NumberByKey("log(x)",strInfo,"="))
		verticalAx = log(V_max) - log(V_min)
	endif
		
	GetWindow $strGraph psize
	return abs(horizontalAx/verticalAx) / abs((V_left - V_right)/(V_top - V_bottom))
end

static function hookClosedPath(STRUCT WMWinHookStruct &s)
	
	DFREF dfr = getDFR()
	SVAR/SDFR=dfr win
	if (cmpstr(s.winname, win))
		return 0
	endif

	// we will use cursor codes to define the mode
	variable Insert    = 18 // insert node
	variable Move      = 21 // move node
	variable Zap       = 19 // delete node
	variable CloseLoop = 33
	variable Idle      = 0
	
	// status definitions
	variable wavedraw  = 0
	variable polydraw  = 1
	variable editnodes = 2
	
	wave/wave/Z refs = dfr:refs
	if (WaveExists(refs) == 0)
		SetWindow $parent(s.winName) hook(hClosedPath)=$""
		return 0
	endif
	NVAR status = dfr:status
	NVAR mode = dfr:mode
	NVAR pointNum = dfr:pointNum
	NVAR polysuccess = dfr:success
	SVAR layer = dfr:layer // when layer is set there is a drawing object associated with the nodes
	
	wave/Z nodes = refs[0]
	wave/Z wx    = refs[1]
	wave/Z wy    = refs[2]
	
	if (!(WaveExists(nodes) && WaveExists(wy) && WaveExists(wx)))
		SetWindow $parent(s.winName) hook(hClosedPath)=$""
		return 0
	endif
	
	string strNodes   = NameOfWave(nodes)
	string strTrace   = NameOfWave(wy)
	string strObjName = StringByKey("name", note(wy), "=") // name for drawing object
	
	string strInfo = TraceInfo(s.winName, strNodes, 0)
	string hAxis   = StringByKey("XAXIS", strInfo)
	string vAxis   = StringByKey("YAXIS", strInfo)
	
	strInfo  = AxisInfo(s.winName, hAxis)
	int logx = NumberByKey("log(x)",strInfo,"=")
	strInfo  = AxisInfo(s.winName, vAxis)
	int logy = NumberByKey("log(x)",strInfo,"=")
		
	switch(s.eventCode)
		case 4: // mousemoved
			if (s.eventmod & 8) // command key
				mode = Zap // delete node
				s.cursorCode = Zap
			elseif (mode == Move) // move node
				s.cursorCode = Move
			else // zap mode will be decided based on cursor position and status
				mode = Idle
				pointNum = NaN
			endif
			
			if (mode == Move) // move node and update spine
				nodes[pointNum][] = {{AxisValFromPixel(s.winName, hAxis, s.mouseloc.h)},{AxisValFromPixel(s.winName, vAxis, s.mouseloc.v)}}
				CatmullRomSpline(nodes, wx, wy, closed=(status==editnodes), aspect=AspectRatio(s.winName, hAxis, vAxis), logx=logx, logy=logy)
				// redraw poly if we move the anchor point
				if (pointnum==0 && strlen(layer))
					Wave2Poly(wx, wy, hAxis, vAxis, layer=layer, name=strObjName)
				endif
			else
				pointNum = NodeUnderCursor(s, nodes)
				
				if (mode != Zap)
					if (numtype(pointNum) == 0)
						if (status == editnodes || pointnum>0 || DimSize(nodes, 0)<=2)
							s.cursorCode = Move // cursor indicates that node can be moved
							mode = Idle // don't enter move mode until mousedown
						elseif (DimSize(nodes, 0) > 2) // ensure that we enclose a finite volume
							s.cursorCode = CloseLoop // cursor indicates that loop can be closed
							mode = CloseLoop
						endif
					else
						pointNum = SegmentUnderCursor(s, wy)
						if (numtype(pointNum) == 0) // cursor over segment
							mode = Insert
						elseif (status != editnodes && MouseInPlotFrame(s, hAxis, vAxis))
							pointNum = DimSize(nodes, 0)
							mode = Insert
						endif

						if (mode == Insert) // in position to insert a point
							mode = Insert
							s.cursorCode = Insert
						endif
					endif
				endif
			endif
			break
		case 3: // mousedown
			
			if (s.eventmod == 17) // right click
				return 0
			endif
			
			if (numtype(pointnum)==0 && mode==Idle) // grab node beneath cursor for moving
				mode = Move
			endif
			
			if (mode == Insert) // insert a node
				InsertPoints/M=0 pointnum, 1, nodes
				nodes[pointnum][0] = AxisValFromPixel(s.winName, hAxis, s.mouseloc.h)
				nodes[pointnum][1] = AxisValFromPixel(s.winName, vAxis, s.mouseloc.v)
				// allow the inserted node to be dragged while the mouse button remains down
				mode = Move
				s.cursorCode = Move
			endif
						
			break
		case 5: // mouseup
			if (mode == Zap && numtype(pointNum) == 0) // delete a node
				DeletePoints/M=0 pointNum, 1, nodes
				
				if (pointnum==0 && strlen(layer))
					polysuccess = 0 // trigger an update of the drawing object
				endif
				
				pointNum = NaN
			elseif (mode == CloseLoop) // status changes from 'draw' to 'edit' after closing the loop
				status = editnodes
				s.cursorCode = Move
			else
				// stop adjusting points after clicking elsewhere in window
				if (mode != Move && status == editnodes)
//					SetWindow $parent(s.winName) hook(hClosedPath)=$""
//					RemoveFromGraph/Z/W=$s.winName $strNodes
					StopClosedPath(s.winname)
					return 0
				endif
				
				// stop moving nodes after mouseup
				mode = Idle

			endif
			
			CatmullRomSpline(nodes, wx, wy, closed=(status==editnodes), aspect=AspectRatio(s.winName, hAxis, vAxis), logx=logx, logy=logy)
			
			if (polysuccess == 0 || (status==polydraw && DimSize(nodes, 0)>1))
				// draw the poly (the drawing object becomes dependent upon wx & wy waves) and revert to wavedraw status on success
				polysuccess = Wave2Poly(wx, wy, hAxis, vAxis, layer=layer, name=strObjName)
				if (polysuccess && (status==polydraw))
					status = wavedraw
				endif
			endif
			
			if (status == editnodes)
				DisplayMask(1)
				AveragePixelsROI($"")
			endif
			
			break
		case 6: //resize
			// even though we're using axis coordinates, if the window aspect ratio changes, the path shape will change
			CatmullRomSpline(nodes, wx, wy, closed=(status==editnodes), aspect=AspectRatio(s.winName, hAxis, vAxis), logx=logx, logy=logy)
			if (status == editnodes)
				DisplayMask(1)
				AveragePixelsROI($"")
			endif
			break
	endswitch
	
	if (s.cursorCode)
		s.doSetCursor = 1
	endif
	
	return 0
end

static function/S parent(string str)
	return ParseFilePath(0, str, "#", 0, 0)
end

static function MouseInPlotFrame(STRUCT WMWinHookStruct &s, string hAxis, string vAxis)
	variable xx = AxisValFromPixel(s.winName, hAxis, s.mouseloc.h)
	variable yy = AxisValFromPixel(s.winName, vAxis, s.mouseloc.v)
	GetAxis/W=$s.winName/Q $hAxis
	variable xmax = max(v_min,v_max)
	variable xmin = min(v_min,v_max)
	GetAxis/W=$s.winName/Q $vAxis
	variable ymax = max(v_min,v_max)
	variable ymin = min(v_min,v_max)
	return (xx == limit(xx, xmin, xmax) && yy == limit(yy, ymin, ymax))
end

static function NodeUnderCursor(STRUCT WMWinHookStruct &s, wave nodes)
	string info = TraceFromPixel(s.mouseloc.h, s.mouseloc.v, "WINDOW:"+s.winName+";DELTAX:4;DELTAY:4;")
	string tname = StringByKey("TRACE", info)
	wave target = TraceNameToWaveRef(s.winName, tname)
	if (WaveRefsEqual(nodes, target))
		return NumberByKey("HITPOINT", info)
	endif
	return NaN
end

static function SegmentUnderCursor(STRUCT WMWinHookStruct &s, wave wy)
	string info = TraceFromPixel(s.mouseloc.h, s.mouseloc.v, "WINDOW:"+s.winName+";DELTAX:4;DELTAY:4;")
	string tname = StringByKey("TRACE", info)
	wave target = TraceNameToWaveRef(s.winName, tname)
	if (WaveRefsEqual(wy, target))
		return ceil(NumberByKey("HITPOINT", info) / 20)
	endif
	return NaN
end

// Create a spline that passes through a series of x-y points,
// optionally looping back to the start point.
// Can be used with ConvexHull to make a loop around a cloud of points.
static function CatmullRomSpline(nodes, wx, wy, [segPoints, alpha, overwrite, closed, name, aspect, logx, logy])
	wave nodes
	wave/Z/D wx, wy
	variable segPoints, alpha
	int overwrite, closed
	string name
	variable aspect, logx, logy
	
	segPoints = ParamIsDefault(segPoints) ? 20 : segPoints
	alpha     = ParamIsDefault(alpha) ? 0.5 : alpha
	overwrite = ParamIsDefault(overwrite) ? 0 : overwrite
	closed    = ParamIsDefault(closed) ? 0 : closed
	name      = SelectString(ParamIsDefault(name), name, NameOfWave(nodes)+"_CR")
	aspect    = ParamIsDefault(aspect) ? 1 : aspect
	logx      = ParamIsDefault(logx) ? 0 : logx
	logy      = ParamIsDefault(logy) ? 0 : logy
		
	variable np = DimSize(nodes, 0)
	
	if (!(WaveExists(wy) && WaveExists(wx)))
		wave/Z/D wx = $name+"_x"
		wave/Z/D wy = $name+"_y"
		if (WaveExists(wx) || WaveExists(wy))
			if (!overwrite)
				DoAlert 1, "overwrite existing wave?"
				if (V_Flag==2)
					return 0
				endif
			endif
			Redimension/N=(0)/D wx, wy
		else
			Make/D/O/N=(0) $name+"_x" /WAVE=wx
			Make/D/O/N=(0) $name+"_y" /WAVE=wy
		endif
	else
		Redimension/N=(0)/D wx, wy
	endif
		
	if (np < 2)
		return 0
	endif
		
	Make/free/N=(np+2,2)/D temp
	temp[1,np][] = nodes[p-1][q]
	
	if (logx == 1)
		temp[1,np][0] = log(temp[p][0])
	elseif (logx == 2) // this is rather pointless. doesn't matter which log base you use.
		temp[1,np][0] = log(temp[p][0])/log(2)
	endif
	if (logy == 1)
		temp[1,np][1] = log(temp[p][1])
	elseif (logy == 2)
		temp[1,np][1] = log(temp[p][1])/log(2)
	endif
	if (aspect != 1)
		temp[1,np][1] *= aspect
	endif
		
	if (closed)
		// check whether nodes form a closed or open loop
		if (nodes[0][0]!=nodes[np-1][0] || nodes[0][1]!=nodes[np-1][1])
			np += 1
			Redimension/N=(np+2,2) temp
			temp[np][] = temp[1][q]
		endif
		
		// we have made sure the nodes wrap around in points 1 ... np,
		// figure out values for points 0 and np+1.
		temp[0][] = temp[np-1][q]
		temp[np+1][] = temp[2][q]
	else // extrapolate beyond first and last node
		temp[0][] = temp[1][q] - (temp[2][q] - temp[1][q])
		temp[np+1][] = temp[np][q] + (temp[np][q] - temp[np-1][q])
	endif
		
	Make/free/N=(segPoints,2) xy_out
	Make/free/N=(4,2) xy_in
	
	variable i
	for (i=1;i<np;i+=1)
		xy_in[][] = temp[i-1+p][q]
		CatmullRomSegment(xy_in, xy_out, alpha)
		SplitWave/free/OREF=refs xy_out
		Concatenate/NP {refs[0]}, wx
		Concatenate/NP {refs[1]}, wy
	endfor

	wx[numpnts(wx)]={temp[np][0]}
	wy[numpnts(wy)]={temp[np][1]}
	
	if (aspect != 1)
		wy /= aspect
	endif
	if (logx == 1)
		wx = 10^wx
	elseif (logx == 2)
		wx = 2^wx
	endif
	if (logy == 1)
		wy = 10^wy
	elseif (logy == 2)
		wy = 2^wy
	endif
end

static function CatmullRomSegment(wave xy_in, wave xy_out, variable alpha)
	// xy_in has x and y values of four control points
	// alpha=0 for standard (uniform) Catmull-Rom spline
	// alpha=0.5 for centripetal Catmull-Rom spline
	// alpha=1 for chordal Catmull-Rom spline
	
	int nPoints = DimSize(xy_out, 0)
	Make/D/free/N=4 T4 = 0
	T4[1,3] = sqrt((xy_in[p][0]-xy_in[p-1][0])^2 + (xy_in[p][1]-xy_in[p-1][1])^2)^alpha + T4[p-1]
	Make/D/free/N=(nPoints, 3, 2) wA
	Make/D/free/N=(nPoints, 2, 2) wB
	Make/D/free/N=(nPoints) w_t
		
	w_t = T4[1] + p/(nPoints)*(T4[2]-T4[1])
	wA = (T4[q+1]-w_t[p]) / (T4[q+1]-T4[q]) * xy_in[q][r] + (w_t[p]-T4[q]) / (T4[q+1]-T4[q]) * xy_in[q+1][r]
	wB = (T4[q+2]-w_t[p]) / (T4[q+2]-T4[q]) * wA[p][q][r] + (w_t[p]-T4[q]) / (T4[q+2]-T4[q]) * wA[p][q+1][r]
	xy_out = (T4[2]-w_t[p])/(T4[2]-T4[1])*wB[p][0][q] + (w_t[p]-T4[1])/(T4[2]-T4[1])*wB[p][1][q]
end

static function GetThisVersion()

	variable procVersion
	// try the quick way
	#if exists("ProcedureVersion") == 3
	procVersion = ProcedureVersion("")
	if (procVersion)
		return 100 * procVersion
	endif
	#endif
	
	int maxLines = 30 // number of lines to search for version pragma
	int refNum, i
	string strHeader = ""
	string strLine = ""
	
	Open/R/Z refnum as FunctionPath("")
	if (refnum == 0)
		return 0
	endif
	for (i=0;i<maxLines;i+=1)
		FReadLine refNum, strLine
		strHeader += strLine
	endfor
	Close refnum
	wave/T ProcText = ListToTextWave(strHeader, "\r")
	string S_Value = "" // workaround for Grep bug for Igor 8
	Grep/Q/E="(?i)^#pragma[\s]*version[\s]*="/LIST/Z ProcText
	if (v_flag != 0)
		return 0
	endif
	s_value = LowerStr(TrimString(s_value, 1))
	sscanf s_value, "#pragma version = %f", procVersion
	ProcVersion = (V_flag!=1 || ProcVersion<=0) ? 0 : ProcVersion

	return 100 * ProcVersion
end


// PNG: width= 90, height= 30
static Picture pInfo
	ASCII85Begin
	M,6r;%14!\!!!!.8Ou6I!!!"&!!!!?#R18/!3BT8GQ7^D&TgHDFAm*iFE_/6AH5;7DfQssEc39jTBQ
	=U#t@KS5u_NKc'hjQ<7_k>^N6#'iu's#@iL6?EN#,SR%ORqWY?g_DQ6ZakX-9m6UF`Y7B5's^*$E%B
	u3<(a:i;5<Ct\*^a)hT'K%0QC]d[0jPV;N@+9cG@fjB`p38^oR<*NngmMP4QXaLkr6j/Tp8k\$ZiA^
	F[J/@;&;!fa/0Y"8EKlO!*92tV8I!509L_n[prdb[0DD]b^77\4Qh+<X22WSL=^Tpl!>%R?bOV1l`F
	.H/kj.Q0YHO!p0Z_o!(t!"2!0>1<o&\&oX/k93#9I:/\ol*"]HFk2"&^kI,Op^Po$LJ:qE[=Zdn*%1
	/%c)$r:,4%r)#UgVCL9-4fb:)2uc00,SCF2g"FKhP:)ciSp&'<om1=hD5F)*"ddDT8%a$Z"blrKB1V
	t"bqBnEcNrlWWRU];Q?b;\6c%+`!0V&SGg^%NRV$e":,J(6+Q,'.qs8WXZ?]HQd?\@;-q_mAJ,]9,i
	ondb3ItHg(Zjc?GW[$h>t<@#!7L*gWDnag1eK/1,*>Cr1P3*D[BX`3`d.+2D9NUG(..rXX^7$1"@82
	!1KE^Ub4(b^NWMsF;Ht6&&;b$b#dun1L:o\#UFM.3gNK`n[l7Cp+c[X/cDaa?O<b.!b=-Pr2.^?1Nd
	)e2#U:Q=,SCE:&8^%*L$dX(UQj\B/Bea\(..rm00Qspp'-Bu;ci@[g2Ab(7h`\ZA=S45.(4!_jMmH)
	qUNK^MN[#3[D>08n\uMdd,!N;U*ru@<Dd(/5&X;O&;3@n7RotRD4?;%HAGk=G+fW.>tiHTqLp0W7=4
	1t\u)fWAJng`QIdQ"8TJr%$paau+6@''?G#H(RO^<AUs^^G[]\cSB@c5aI;=JigY9^3e#1T9SXPucg
	hq>mG`bOMiSe2FC2.Jjp@_T>`piIOWDnc;k%;&=gQn1M0p,[l:OIIe[!*"QkrNrEAXG*0N(r07]$fZ
	h/$5=fV3Tb,=$\psV/^EJ)4IFB]PEW'\qbg(SsMlA:JXaHWi@OR/f1s5k:MsaPoMc&J--#d&-rCGg1
	@0&:")(CjpXC94`bia)`#WJ(G?agO/a)!%!Za'1b>,JjlN'!=]&V5U;CI@Et9%o2-h+N"!1EAB5`,M
	Th&'lHW;8P&FJlPBUVCZ77C(9D;1qiXUW&&O5KD=a,V0/D%4B*&."qC)Xk?1%Zs:M/7mre"!%<ET2<
	bOD<cB%HGLYk/]p$:.Z,bn[7sj`?B9ZXXV6Rq]_Pti2gd\J3f(i:R&5_pm4KmDXm$aS/B^J!/^)#uG
	%6sfYult1cH[_1nRTfUip]H>d<$0>4r9NH8sWQ9)djkQAg3E#:]C\]o'uTN1rJBf7mE:cKPfJ<US@o
	sfC=1##U)g+pYUIm80gMI'&gcc'97IQ?8%%Ad+-[#D;.]AGGtPU1e&m.iUpnuW^=]Xm<&%d+jRgmR6
	q<!"K4)c%%V>s6.boU<k:W9@PN6o+0)@'G&kQ@q=8),*BN8CpR@_H@2O[8@=!9Ifs"n;A,h79HFe;'
	<AQ<<*SW3\4hYn'7c+!LCK<HgOtdE6\7T^B_C4`os.\g?L1-<OUon"8QFD=fRPcBo,tYs-%!5e<il-
	gK-hM%Q-;cW@=dI0o(g+V`I[7Hll^gmF1C8U`9;n:SDf?'4kXX)YA6CN6o]r]qT!s1+^3pK)B6pt_B
	a+O@`IRg8=:@XkJL"[Wg"k-TK%RqsK*5:)MTX360]U/4Wi@77;Jt3uP7<CJ.#Pn:q"VM]0-o&(8.o9
	`MEo2(5'MS@J+Rq@q<5kKI!A\?`c`\j5Xs.C`^\u^];Gr<!]0NL(T`^;0>7#+XNrGfbj/DFR`<:aKO
	an5L'f'H$C6<N=ieH_Ns:8r/-F>E>*(o2.ET1%BpgNq9:F_:'kIT8!)),C3GlcmYHU;j&0aK<:WMel
	3Ze-Om%!i`\)#_:DlKDdZeWg@P_C[O5M'iUB8>_-2HE;O1>jM;$8=V+SNHPXHb;/UA2:]pC+19D\Bd
	qtPsUXEa5qKga#B2SZ=.SRp%A!RdqL(!Ea7bt3ql@F;Uj=M08??L+[=OW75>HDQ5.GO:69br-505tC
	KaJC@jCMg**JRJQascSIH0iWNK&pUTkRi35A[F<CY2`i#$#_f\:6RcLI9_:V"&,4[]HN!OfQlP*Qtb
	VGkuA?F_!6F<)ocgU#D#V(':e98CFiHpC[p2PjB,D#;1";i6n,<g%nVGOf<>lBjHXs8BA(fD;DdgE&
	_+N0,P*1$jPnaB'W+c:JFR2daE`AbT2\$"\Atc30k=3]2,.R>E*Nf"c^j@=F?-EhKuZgDlI9;bg[c+
	#@$`/2$kB@MhsTDp>G]B0oZ#`%fgeAhHcB?i5($LInM,P9MS?gJ,auQDYK=UG;=]pX[t0`Ddn+h)@5
	*f*'#BK+[7+e+25ad[WOTm!ShS%o^pU.1GgqQgF=j%Y7Q;ZTcp"RlDZ'E)=kG_qYSr#Wb]AY@*-(nJ
	s#A#q"YFMfsgRh]Kt<,gI6=12lg#`?;TLp_*KtthTY]\gR%W:!I8"sroVaPFBo;RFjF5*7K$9'/M"a
	&=8b8<:r],W]a;6"MF%U=g2!m3l-^kmCMU]X>\[CWf<8QErql0Iq<+A?>3\a$'Y&&M_Vt+amGf@imo
	bQW3d]\44-GFK,M+U?MapjQn$rZ2MO]A366&(1N))Y5K\G-`9XF4l;%/:LK#,8h*dQfYG3mlOSIcZr
	>bLn*pC6ZP>8Kg\bG_@CKr2I%"\bg,9T7gnab8=Dm!=-WrDiH(kdbct`7o4FNN0!X<E0c>,]?+,R:L
	D]PZ0,$>?`&92Jr:c#b4?2.q.C==rdOGT-GEe&/PNlY926R6do>OS3)p0O\R3Y8&I`m-S_GCD(TPUp
	#Oh0$f?=CMoA(S7gE=*>mMVCq0YVL.YYcK,Y=`@\X=h:G[ec2^3h=YF;C/g`2VT*c?qXpKj]]U?cVq
	fC?DNujYDHgQ#<&Y/Z3l"'p#PjIED$6^p\*"iPGcpqYD$G(gq`=S8``85)IVK,7k,W='%mZ6N^%W9s
	ZH4Ib4'RZ77*WgpE7az8OZBBY!QNJ
	ASCII85End
end

// PNG: width= 90, height= 30
static Picture pCog
	ASCII85Begin
	M,6r;%14!\!!!!.8Ou6I!!!"&!!!!?#R18/!3BT8GQ7^D&TgHDFAm*iFE_/6AH5;7DfQssEc39jTBQ
	=U%adj95u_NKeXCo&'5*oW5;QerS2ddE(FP;'6Ql#oBiuS1dW/'TBu9uWoJd,<?@,Ku[OT_*C6>N&V
	5fgKO-Z!0m]BHZFE$kSY/b+).^UIV80KE-KEW"FQ9_,eF^bT7+UPEE3;1R;IlL+VJZGMtc/)57aYSB
	X^5kLMq7Sp9q0YBG@U214ph]9b)&eA4@(!#Ip%g0/F0_UH>\E'7M.=ui]mI6`m'G"T)B'OA[IVmQDm
	=?Eat`-Qd[5bR1c#F]`cT(Se*XgJen[DI=<Y8YF7N!d-X+iY2pq;SVDt;siWS_bs$!E]WT^pRhs]aH
	M%csq\akmJ4tMo<HGU1WqT,5cib%XrHM`>*]/&U:](RQ7QDj\u&#cMNi8>`?8;-?rCcXX>+1^gW10F
	n!h-2G-M;R\a+TD2op'?`R]"%W!DpJmO],i(@-(/F?,;L7Vkt%YOa,f\8+CV@lHVOEMgMZPnh$6>!V
	MTYBm^8f[O,?Z$2MnH6.T'JWSM4HtSissRo-):4d:ecoe5Tn^(gUEQm+o94@;L(/[91%aXk:!pP;mm
	\kh$s.7qbe%=-p1eBtDs*CHp:+$CUY\0A,jM0:4J2&pY-HWBG?nb`"BE/M-#*)+E?I*C/(r;J]APNh
	3!Ea<(u)`o?0R`ma=QV<n?GV/s3:I0Wf2_M0p@:__T%OEl+sL@10K8&ViQgR(0Q3qMLYA':/iba:,;
	]Y$@ACMV&9b[fD4A`Vq5+A!37VD0na`;0#fWNWKq#f5N>Mt)$S['[2:?=(p2$Q$$NX_cXoJ`iVOcHm
	Rb+"_b#*b4@tp)Xq9r*1_<^IVlpMJ=kE>MhiHa2]]q9<d*4(lA_8$4ej2NM5Z!#`oc=+Ttk-]%D5"O
	Yiu,o$V/I<=@2fN3Ds,PNfIEnqn6C?^[OYDs4q2k*s6TFu+@1>SKUmdko@B5>Pp)-]8`l_Ig,/1c.T
	K'Z+asa)qDc*mqZAKmijlOd;;&H$MEMWY1:\q<G#aaNVlho?TWCGL35!G658MH$RpQ,/[:S#==eP-@
	T+s%'h-7&.0)\eW6j@1gNW'FYlgRIid1g1dP0.MtL)"o@*4A+B&XU\9bRSJWg?B!keI%b6T7FS'?W(
	@7j-a-n[,Adkh%U((6"oN9G"iRV$rmb0"2lqXk08!N&V_b13Oo*tc)h=[L]E@W<ihr:]%Dbs-*cCbc
	T^`<b81D1(d_gue7JX)rMl-Q!ag!0.a4mbCL5'MQH2X<p3`1nA#69QNiWDnN[J^:kIm`JPCXo#W8lo
	?KFN_dOMp#7s\i!;q:1Vb`q^Za5j'0Sg8ALVn\tm:OM*.G/IF;=K4S+O/0U]^`u\O,i'69Y*0'f,SH
	:Kp)mH3.EQ3hDf%?f;W[LbJbR7R5Lb$8U4I7,[8ZM7fU7H5(>6BlSlr1cG6"263q!"YH.\`>aLYN(!
	e3hZYm666$Mq_c(_GHO?%CE(rnI-UV=I6M\e$%CXt$`9q"IB8d`/4e)0&Dcf`43oobf6MqdV?d>\73
	X/dI(2jZ+#[Nt'sQg5KRFl_^rEVX>cZ5h3g1gR#*IhTZf+KqmsVrW[`UcE6>MM*N0fMcT[S!:h'=ju
	EaL@5>OfU]WXd+d/JISLYCQZ_BPkB$IiAYUBq1l^ecC4a8EYJ'WJ,pak5V59k6$F23m\(d</D&W$.c
	&:n64N(\`j0%h;m7iB=j)(R^kh8BUCfn<.JP[26K1F7\>0JOc56jWCOX(6k=i$m^+A%<G5QZ.i$e8]
	01cU/k$R?&@c[1P\L>tK[OkSMmQ7lT_puI&4&#-'R99q+p;3\Sn`Ic2G%kDj/"N6oB<E0?Z6Kl"@,Y
	?4P5G,Nu\q^LTs(Vj(g3YCoZEl;W9!T)>ePG,S!10n_Y\rP(FBr90.J3;jVP^MYp<-ML6>(=)*bFsB
	&uWX-VWg'<P8G@$+\A?J1@G'F,/Z.<q/;.,=8I?;gFp>>;IjEQPE_:7`.R+3bElA@DB6<k@ksJ9lg(
	=CVM=g<G(^E#SiiFHZ8.qF-^ppkE&\[U*_);<'Lfk*Fq]^#W339=Q'IWpj"KgUWU]1,ETW<&G^OtOH
	MkGSX1r0@A!=Gd&>m2;$s(nF8b!K?8R`eZjO8V38_I$<1AcnP!)B*`T#3.X=hE[s8PMoFf*/H2@32u
	N'Len`GUk?n:K*@=9gMMi1ES9gF7flQcCD`2n^,h:`S5=GNQ^G#@^/a:?]W`PV50n4Xue>QVk8E1=]
	lWKB?uV(SiXjL_hVC,G-+W-bHd)[C`^u(BPM:VV58ltJcZ8d$CEhp-Ek)Qb.'^'f)4eT`-]7*:O[1>
	WO?>HRRZ3%&Aum4gNS.jU=.^S*B#H\'0H3Z)tG>O;)WW1_IO09[>@PI4Y3R7J0]^!UgQhj+u1-,O_#
	4q7aj2?Y5.V]ph=D*J`g8?n%JH:8P)K%MLrfVTs(X1]A:d+mFtdNBG"";'8siHNZC4&bKJq`%mNb7r
	SW;=`2-+n=L)HDOsFHoS$CX_6m<3W758iRSt7"9?7u`s%5^"&NsfV]&fKiR),nod"F>!-ZZj257RH'
	BCpkU2>pgE:BYW?=m,Fa:(R[)N0g+8U_d1?h6?87.J?1.S8QM+N.?c/5H^[K9Qq/KSe*09PFi*)kOs
	Cpa8151hB![K\`b9:/2t#`0h)TQGGW^_mOCaj@jCA@uU*q95,uIW@7!X&<O\"P/_=YGg"!\PP(fcMs
	XX8]B01I/5GVm/oHal.qLkA^e?r`*cq5<6cMYpEN]i/%/Vl(p'^fIMd8C\oHLnT1!)7oQm]mKeID*1
	Nn<=:2#ZkEe#YpHHH,[1j)T%7I4;@/)4cu_Q1)PbZM:[@i"UEI%;r>p03XoVZ_RkU=+$'7#=USG/bL
	8-+m<=>h,iqN=A8kNQ3E0-ce+TlUO7L$\:&5CW07\^Y5(=Lpj3bn5fXf]+hD?H]7Wq?&[-c"7hNK0#
	/)B'Mj<HT8ma3CurC,*bh\'aHR7Z[QbX-ZmAk800m):lpO:?P+(!;TY'R\j"AmjWGZ^'4n^bf?W3J@
	>&NBK1HqWkVhSJ@2:%U$9,hq:d,G1cCmI-TcrP\J*Ut[2@6?BcK3XN6]^DH?sm>]m;PWk0+t]M3*pb
	_i5ToaNr1&dko4ib1O7G-^#a3R58IWd+6c;6ULrU<E06&]A7B"H::^+p=jM"Cht@E-\k9W-F%RN7[f
	g9`s&hll-^kfa*7KZfMd!Yc("^(?tb@(G_h8BF>IEA!<RhTlND,!db&r!Y3m>2$M&6dHhJmnR;"(TN
	7k9deNFM:rt_%M:_\bI5MO2h<A0N#R:ZSrC"&ps\iY*%&:=-;@IrX+"G9!l_&sOI?=_'7)$hsk)[Og
	CfZ9V$@mNB]AS#G_>V6^Z_/)$iF?19\*_+U8'Lh!@O$@74\ogtP<K2K,l#,B;0e6NJ*#oS35C1Gqa?
	[/#EMYbdp\'g083t^Hm.OC']=?TCh[+\@'/CD^$lb9k>s0TnKIaqh4fI!&lDq*\4*U*,*??/2AnId;
	.P@%q^Y_gV7L#<Y@CP!Nm,EKn=&i8<r@!PTa5]H_PQ](fBqn;iMUE,Pl5E-<#!*W9G0Gifi6\]*<o4
	FnfrX,WF^_\F$1nF]^fU-bLO!=bm%5lG.k3$IWMqUu'c@l,R*B4I#7$5RGsBAo;S"q!W&r%8C2/"PK
	bsa<4<J73edT0;DW#W4)GM@uFDONL"9R+_N^([T!(k&'aJc*VLHV&^R;!(_#4Z$g7@#[rnEd4b][m6
	"s2Cf6FX#Yth+![-Bc9;DCc35!#ZOe]>R5l%A3s9r*"E3cZ^HAq!@!X3ZLR*.A7oQ8om.\p[kV)RYJ
	YS_VJ,ke(!91A):`eg`H3<A0s%C//=2RBq,pC2:S]*\P@U_Oa*3.T[g!Hf.uI$^Z574:Ig+a&Rh&b)
	g:i!IBPVCY]Y$?Mc-nM/==coe'#A=jP*M;$C2,4.LP+[KAA[:ZiG]VW6ipmf;5gRtUogbYmG#,M=Y+
	23\;kLm>:;.qMlKqn)uClJ![.i',6Vo?VRu"<5C39QHia.rIUXNcq/492/E<t4:q=5j]#0#BT^2C8R
	r=7NOaA&EGCimE'I"(o(b72sE7cRPmWBnj]tHh/;(=(Hs(iTH$*GN/REC6P4"-OQ)1Z*S9OlNXqYG,
	/qFo"%Xt5WcH)DMDo_@'gn2mp\l)\(b!Y0Pa!2n,Nj%-R@Yjn?WT$E#t(FUbj=.R08ON,:0qYL%:/M
	0]<Q1)2)G':0@s*h8ZZ<4MLPu4'A3d&SI6l9j+cCI%0ltDh?G(&13<0N\Q=?t??;p9Q8$59_n3Rli=
	5b`N"1`$#>;[JNr+$!*^fli%qGlHBaGd#r_gndbH0<a<pR0sE5L0:j)I_t)\EH/7Wqt8QJMd<r<&WK
	8J3cuoH9hij#22_bS-?/1q+bUC@(DjDc_1Dg*LCYK([C$_m"OB=44C54XF6CiRHM)#JSik-Qi#lgdX
	C:AAV;n0(-%L_0m!T,"d5MVG@HhTKZ87K.p%WH^Xh3kCpXcTY<@p]%p!H!PcGm8Ma`dWI-i:%O`.@A
	\EMhGlp>Rk7O<G2m`*r,h[u\8;4r,bUaQp%EDN*J]D4B1hFXuppq^tpModARV5%<QlNN?L%hAEkIlW
	/#`^]Bs#-d.f-97RH2#l<o@ZXZ&T?iRH-<%N9SU+'so7AAgW2mil0fWaM)A,6@)4Rp@WFC0@Y,uIN:
	5uCJkMPAJFd6VVd/UR6[rI=?0fVgq/kI,s^(YARDN54WD\PD#"bV<C5XKS<`]Fql#m@)ul]O#Nn]-2
	[lB);<HM,sPARkJs:;d4_S!/nh?qGe7@I:AsnMi7DjM_D$2XW>fsY^ZQI8$;`nm!f"mg^^mq'BNs/!
	!!!j78?7R6=>B
	ASCII85End
end

// PNG: width= 90, height= 30
static Picture pHelp
	ASCII85Begin
	M,6r;%14!\!!!!.8Ou6I!!!"&!!!!?#R18/!3BT8GQ7^D&TgHDFAm*iFE_/6AH5;7DfQssEc39jTBQ
	=U%&ule5u_NK]pa@g']g@U^Upt%GFQ2iG6aJpJmdFX78;,cVIC:7E<8Td:b6cN,?9UjVf\f%WZ;[/V
	%KP#JO*M&WN?THKK9cU)`=`H0[O9,)L8;-_'G*Bak\80]?@nUmHm(8j1N'qpQ@[GqTSkLm[iV;4l?5
	9p34hHca--pCc9?`a1`N@?0>MW\`C_cK87>p!!)X0(AW7k(<k1!Z6U5Zb%=PW#*`9'?8(R.l9\9B%n
	9#Xi,CY"\p?O1oCK4p_#dS:<['Ueqmu'!DdI1i5jrP3+p\-^L%B)Rl&&6R@eFq>=L[qbCBEtmmNr/L
	*e`#,Pc(S2@J/p9W2QZ)-J]W[Rki??laDuDX[@&P,(^^j"-[%#SONo\V1=G'!!%70KV'(>nd;T\r9s
	nW!!']u`J45Y;-RJO#^h#IS8@)s?-6E37K\jr"J&9mNoQd1W?So9&&7!4QcfPj`<*6p)q!4%Z!kNSb
	ab*S0C)+!-4P&qo6Jd<2$QkkZlF^m=^($s8m)tsn,tIZn\F/q!JI,[b(u%<c'JV4aX<^Bf%*k%J6&^
	>;,IqtHG)FHLBE0]KCF+J!.aqBH9Cu\K<fV!+1qo2IdrI^jN)fkTP5;J-4NXp%T0D`+g<s/YW;Xf"@
	\,T!Yqbd?E6KnN%F4r'*)'g*uD%cKYq4thknG`:Fo$Eh;$dZNZ:%ld4Hbsa^bHA79m^g0PiWdCc346
	JH/)8KM&)(k36QF'1-*7>&j)l`JADS2'.,2Z*_)fH$KMmMP'\-](',%?j1@)>F]S9TO0p.T!W[#6cg
	fTP-7;t(+Zf_JqtSn,aMH!n\';3U2ZD&S;P.=h1rnGXk<CB$fKN,kd+9PO4u%TJ&Hu4&4Crs9Y@_nH
	"MR&l)1>b[1u<1&g\hMaNi(kl1-rkL5)pKfeO6U'VM\7&0/`!A7PfqQOE]am/tCKELH9=]@#NnO'jY
	tqYGLh>IH)^mct,(P9&uXOhUj'bDC#sD;2^*qE?dh;P68RpF'`l^&V357%l%naPeh(A($.glRQtugt
	ZGl%$*W+%Ia6#,81..8g^]id,Kck$t@][$aBOjq/OasMhNuG2M,LC'G_J("Z[P4_r5QBmFntXjd',6
	lZt\G^OJI;7Rj%OX5iIM#6uCX!]FoLK`(amh5S'MpXVq/O#EQ=C>KI[/s=orf,T\.V5"J]r\=n:<P^
	dbK.Jkf0au0o*Zjc5<u!+J$%sur)3hjN3<hR:W<>kX3+#G"^cXsdL/hY`gSKQ2hQ^Pq(l!UpCdqH+(
	.;#abis6/fk6q:$aRh5_>*eM9NS:6RPelrZ>?$aa)Nf&c!V@m]6<R6e=FIgEcWdq5MWd4Wd(EGH-(t
	7'6n3kpg)VSEUsmM3$iUNPR)sTB+NfY$cV71VX=4n!'f7b&o2]5j7tq;=*jFX8d1ff/P>NYa\kIC.H
	0D<]`$GJ&fmmVZ]hoTDId<i+X\WBThm/_*d[#n'GhQY]#TjRa0oj>Y^?+^*&qn&>%qpqLF8^A1;Q1&
	-O1QL#&^mh-Vg1nFHc5Yg_^Fm5Mj%+,Mn_\Yol9dGTnRi%0.uNj'\-e(in+;$'enZ/=ITkLsojIs'f
	5J4*Sf)rJ-0)p,hFVP)c6^;1hlKdF&nZ-FC.eJ7Vk#%#29.N7pubh'DZ93h9r&DjT2'SNDUuSfXk:T
	hm!T6a,*oAQcCI[)@&bnCG;H\^W#'+J+VfO[ZNjXF'@+Pe?YI;kq8O!7WJ6/>GbW`_[MnT91E/eol'
	bTm^=u,IJJ5lD.poMum`/$SP;Lj,6doEk."\m=nfH#0X:Yfbb'MOIr-edqrknBHJ'.M!5h;8]P4A(2
	C5q?EY.TLP&:@>Ph$?Yg`,3D[$dC9?0\Lot/&E.W?j`_I5L)S`l("cMF\gETWuSY0H>CU9[07$'*X?
	.ECCH8hd=RSR5*cX/cU:_gpc,KYucE`q+RG\$$<TcCB`,F1^8,XK7BFNf9bRLZ:?J%"ljN:<;Dmlr<
	<hnCr)p-KUd:j9K#ZR[0<)\a0F9.RDN?6IID#?^a.u`5-q,0f8$Rlf7RA:d.C+<VOe$lIo%e;iX5o1
	`08MhTJ)r/NmJY-DDFd-.d!K_Rt0c0JkU#f>m(GTtJ#XKnY4Yh:U@^b&M*`#/`.YmHm=QB$'%\qX#r
	@k9m_?!'^:&'rj398\D*jgar,Hmr+Pss6/[R]X5WiKS!V5MLl\*=I@$r(oGZ,`<UmS89I+G&ERttG\
	?uGTYJ4QI94'p5&O$_NuCGa:bXG:/cc`G"U[\/_&o4D^R9qQ.&+2[(Oj,+Wf%(,h7@a%l,]TuYHJJ]
	c-7[nR\D2Y'FNK/2Is\:pn`I55s*LoL%75rm!\:SWNQ!$7lE$\oEiU0_1I2l(qmg9mW>fI686_Rj.5
	jeJn6.1(2"3)s(MJI+9@R=1DURoE!UFZS]K(#"X+#C.&IfNW]P,?eu`-]QAr!;/J\^(*&mqjc^33aT
	ho-+@KOjAb]rUUj[oC9NQ;ek29;O&*`'Q"/";k+Tg7'`6o+_GJCHppi2q`P`s):,hk.sAA13FNUdVO
	CqH6RTVp:'Y5Q5n4OSA+pK<_\dBuc<WYCcMO?Q@,LgL<7c;$W^J0;HuM+9W5^S^q4aWOpsd@r>fnk2
	So.FRL?T3's"<s%-?!6l<9^F=LQcfaIk'F`fAVN=fHg&+@:55-A/$,0Q;a!e\;h*Lf(XaXKL[!4-!c
	aqqcW-G6]?LB/PBBUY6(jb9r[dt)"0%,6E?I7ZC,@h&U9d0N'0B"I*cP&=W6+,5Q+jm*N\UX=lkGu&
	=k,OT/8_;BQ/7_Ej-C_!.H&/7RV2C0;(69I-DBB1!=ngL'RTsV0@Thl"/7EIHaEf0:a;]5Cc66VSZQ
	6I"!Xu@+C"F9t&6)h$QNJ4Z`l)J2%pja/;XcRC:V`\qo7%>:kj303LOY7gGU9[=BWi@Q<4J,J?\$`W
	[j`9>Hfs<YZk"J7Y4O9e.Z4_2'41CGG!H$J1+c^co`cG$^kE[l+]TgZQ_ZQAl-Ye&1(foL%GJNZubA
	I=\jNa+r25c`$-\_d!-SF>`c$*"bnA5Pd&@G[NL1-s#NT^QJE4K3ch;$eqe%2otU'SNkO0)Wq;?3TU
	rR^1M8Y>J.e=2WqS,EtAVsr+7Ye3:5n)T/gm&69l`j'CTj4Lg;;));SI-5M&;0PNH(V+qhEd59(q77
	:UfXZ&NG4+.BMM_c:\fMCIl=*8Wl$0*1n"*V,VrbVTeamc>-QOiL!'a[j8ho<N"<e&.<_R:uIoHWYJ
	V2XkU$7DS=[%EK"=$\$kMQ0&ZlWD`GVR/k$j8kX2:&F1\IrM\).p2mX&c?cjWb$4(kt>!Fg\$7MYDA
	UHms%/2@dgf]sf_>hK$8KIS*p\IF@UM*pf19J:qKEM*`+C1)HeU%mj[5%,n0(A&?QmTqML1]D_'#VU
	*PTY.VJl],a[1oi.[R2gCCs)EWsbZM?gV]Lt.iOP5c/>+CgY](/@X'N)9IE>TW8![nBi9rGA-g7>Hp
	7+aXH[X[R9gD:QXZ-eK.]oBqplJdKII8NiR<EiZXD*$*VkL$%p@CP!"\(botnDUa&J]RXWD/Cjs8kK
	_<Vt(nb'K\+MDAp)9hWgu3=Kl-_8`3<f3W%U'e"ljrNP(a<ThrgW8>ReE0a-h(!L0!@m!e[%Zs0JWq
	t?0?mV_+Ij[j)-=03DO_eQTRXK2hGf<1D3EY\ZChfAPHFN"fO]NbF:l!M_*b:=[),oo-29QT>^V:Jn
	:-m.rqHKcs0/5Y;,CW^2$p-3Z_&F&EBJB;>bhJ?i+bO_rAXf/4u_28sXAEi7<jcOgQrqM6=VYgP@_8
	H64h2ZJOlBs?ebm(55\!.MW!DM.KJ-<m>:brccmn<T3VCVWDrT)dYhZZGG0Gk5H\]2"fJq/]7#nTGM
	d*Rkr\k#n?2".f$$2_u>^Ei3E4V/;GBQFSYB4Z]qWhhRd8-hNf"q^:7iL-29.seIBTbbt(s41u^CK<
	qq#AJ*pLEHlLC:2KZR-9u#'FI,;"9\c,GW[%O/;p>tQ_L@Sk*gC1'^rD[Ng<&QqT8!\!;\90X=aNKD
	r`TCVbt6obG&O5$4H$sZ/XWa124<T4?Q;4T0A86rdItO<p@AD3h6a-P6iN]-!7Zr+m.(@HX1ct2t-b
	HR`A77"JcKsz8OZBBY!QNJ
	ASCII85End
end